#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) scope beforelist** (i.e., low-level
dataclass aggregating all metadata required to manage the beforelist automating
decorator positioning for lexical scopes recursively visited by AST
transformers).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeConfParamException
from beartype.typing import (
    TYPE_CHECKING,
    ChainMap as ChainMapHint,
    Optional,
)
from beartype._cave._cavemap import NoneTypeOr
from beartype._data.conf.dataconfplace import (
    DECOR_HOSTILE_ATTR_NAME_TRIE,
    BeartypeDecorPlaceTrie,
    BeartypeDecorPlaceSubtrie,
)
from beartype._data.typing.datatyping import (
    FrozenSetStrs,
    TypeException,
)
from beartype._data.typing.datatypingport import TypeIs
from beartype._util.kind.maplike.utilmapfrozen import FrozenDict
from collections import ChainMap

# ....................{ HINTS                              }....................
ScopedAttrNameTrie = ChainMapHint[str, BeartypeDecorPlaceSubtrie]
'''
PEP-compliant recursive alias matching a **beforelist imported attribute name
trie** (i.e., recursive tree structure whose nodes are the unqualified basenames
of problematic third-party attributes imported into a scope of the currently
visited module, defined as a chain map mapping from strings to either the
:data:`None` singleton placeholder signifying a terminal leaf node *or* yet
another such recursively nested dictionary).
'''

# ....................{ CLASSES                            }....................
#FIXME: Unit test us up, please. *sigh*
class BeartypeNodeScopeBeforelist(object):
    '''
    Beartype **abstract syntax tree (AST) scoped beforelist** (i.e., low-level
    dataclass aggregating all metadata required to manage the beforelist
    automating decorator positioning for lexical scopes recursively visited by
    AST transformers).

    Attributes
    ----------
    schema_attr_basename_trie : BeartypeDecorPlaceTrie
        **Source attribute name trie schema** (i.e., recursive tree structure
        whose nodes are the unqualified basenames of third-party attributes that
        *could* possibly be imported into a scope of the currently visited
        module such that these attributes are either themselves
        decorator-hostile decorators *or* submodules transitively defining
        decorator-hostile decorators).

        This tree structure is internally consumed as input by the higher-level
        :class:`beartype.claw._ast._kind.clawastimport.BeartypeNodeTransformerImportMixin`
        transformer. This immutable tree structure is produced by dynamically
        merging the contents of the immutable
        :data:`.DECOR_HOSTILE_ATTR_NAME_TRIE` singleton with those of the
        immutable
        :attr:`beartype.BeartypeConf.claw_beforelist_decor_hostile_func_trie`
        configuration option.

        This tree structure is implemented as a frozen dictionary mapping from
        the unqualified basename of each third-party (sub)package and
        (sub)module transitively defining one or more decorator-hostile
        decorator functions to either a frozen set of the unqualified basenames
        of those functions *or* yet another such recursively nested frozen
        dictionary.
    schema_package_names : frozenset[str]
        **Source module name schema** (i.e., frozen set of the fully-qualified
        names of all root packages and modules transitively defining one or more
        decorator-hostile decorators).
    scoped_attr_basename_trie : Optional[ScopedAttrNameTrie]
        **Target scoped attribute name trie** (i.e., recursive tree structure
        whose nodes are the unqualified basenames of third-party attributes
        imported into a scope of the currently visited module such that these
        attributes are either themselves decorator-hostile decorators *or*
        submodules, types, or instances transitively defining decorator-hostile
        decorators).

        This tree structure is initially :data:`None`, in which case external
        callers are expected to explicitly initialize this attribute to the
        empty dictionary in a just-in-time (JIT) manner as needed.

        This tree structure is internally produced as output by the higher-level
        :class:`beartype.claw._ast._kind.clawastimport.BeartypeNodeTransformerImportMixin`
        transformer, serving as the glue between:

        * That transformer's visitor methods, which initially parse import
          statements importing decorator-hostile decorators into a scope of the
          currently visited module.
        * That transformer's decorator methods, which subsequently produce AST
          nodes injecting the :func:`beartype.beartype` decorator into chains of
          possibly decorator-hostile decorators decorating types and callables.

        This tree structure is implemented as a chain map mapping from strings
        to either the :data:`None` singleton placeholder signifying a leaf node
        terminating this recursion *or* yet another such recursively nested
        dictionary. This tree structure is intentionally implemented as a chain
        map of one or more dictionaries rather than as merely one dictionary to
        transparently support both:

        * Global lexical scopes, which do *not* require a chain map.
        * Local lexical scopes, which require a chain map chaining imports
          across all nested local and global scopes (in that order from most to
          least nested).
    '''

    # ..................{ CLASS VARIABLES                    }..................
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Subclasses declaring uniquely subclass-specific instance
    # variables *MUST* additionally slot those variables. Subclasses violating
    # this constraint will be usable but unslotted, which defeats our purposes.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Slot all instance variables defined on this object to reduce the costs of
    # both reading and writing these variables by approximately ~10%.
    __slots__ = (
        'schema_attr_basename_trie',
        'schema_package_names',
        'scoped_attr_basename_trie',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        schema_attr_basename_trie: BeartypeDecorPlaceTrie
        schema_package_names: FrozenSetStrs
        scoped_attr_basename_trie: Optional[ScopedAttrNameTrie]

    # ....................{ INITIALIZERS                   }....................
    def __init__(
        self,

        # Mandatory parameters.
        schema_attr_basename_trie: BeartypeDecorPlaceTrie,

        # Optional parameters.
        schema_package_names: Optional[FrozenSetStrs] = None,
        scoped_attr_basename_trie: Optional[ScopedAttrNameTrie] = None,
        is_validate: bool = True,
    ) -> None:
        '''
        Initialize this scoped beforelist.

        Parameters
        ----------
        schema_attr_basename_trie : BeartypeDecorPlaceTrie
            **Source attribute name trie schema.** See the class docstring.
        schema_package_names : Optional[FrozenSetStrs], default: None
            **Source module name schema.** See the class docstring.
            Defaults to :data:`None`, in which case this frozen set is
            implicitly constructed as the union of all keys of the passed
            decorator function and method beforelists.
        scoped_attr_basename_trie : Optional[ScopedAttrNameTrie], default: None
            **Target scoped attribute name trie.** See the class docstring.
            Defaults to :data:`None`, in which case callers explicitly
            initialize this trie in a just-in-time (JIT) manner as needed.
        is_validate : bool, default: False
            Either:

            * :data:`False`, in which case this method avoids recursively
              re-validating the contents of all passed data structures for
              efficiency. This should only be passed when this scoped beforelist
              is being constructed by a call to the :meth:`.permute` method.
            * :data:`True`, in which case this method recursively validates the
              contents of these data structures for safety.

            Defaults to :data:`True`.
        '''

        # ....................{ DEFAULTS                   }....................
        # If the caller passed *NO* decorator-hostile module names, initialize
        # this frozen set as the union of all keys of these decorator function
        # and method beforelists.
        if schema_package_names is None:
            schema_package_names = frozenset(
                schema_attr_basename_trie.keys())
            # print(f'Defaulted "schema_package_names" to {schema_package_names}.')
        # Else, the caller passed decorator-hostile module names. In either
        # case, these names are now defined.

        # ....................{ VALIDATE                   }....................
        # Shallowly type-check these data structures.
        assert isinstance(schema_attr_basename_trie, FrozenDict), (
            f'{repr(schema_package_names)} not frozen dictionary.')
        assert isinstance(schema_package_names, frozenset), (
            f'{repr(schema_package_names)} not frozen set.')
        assert isinstance(scoped_attr_basename_trie, NoneTypeOr[ChainMap]), (
            f'{repr(scoped_attr_basename_trie)} neither chain map nor "None".')
        assert isinstance(is_validate, bool), (
            f'{repr(is_validate)} not boolean.')

        #FIXME: Additionally recursively validate the contents of:
        #* "scoped_attr_basename_trie".
        #
        #We can't be bothered at the moment. One catastrophe at a time! *shrug*

        # If recursively validating the contents of these data structures...
        if is_validate:
            # If this parameter is *NOT* a valid decorator function beforelist,
            # raise an exception.
            die_unless_decor_hostile_func_trie(schema_attr_basename_trie)
            # Else, this parameter is a valid decorator function beforelist.

            # Deeply type-check the contents of these data structures.
            assert all(
                isinstance(schema_package_name, str)
                for schema_package_name in schema_package_names
            ), (
                f'{repr(schema_package_names)} not frozen set of strings.')
        # Else, the contents of these data structures are assumed to be valid.

        # ....................{ CLASSIFY                   }....................
        # Classify all passed parameters.
        self.schema_attr_basename_trie = schema_attr_basename_trie
        self.schema_package_names = schema_package_names
        self.scoped_attr_basename_trie = scoped_attr_basename_trie

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:

        return '\n'.join((
            f'{self.__class__.__name__}(\n',
            f'    schema_attr_basename_trie={repr(self.schema_attr_basename_trie)},\n',
            f'    schema_package_names={repr(self.schema_package_names)},\n',
            f'    scoped_attr_basename_trie={repr(self.scoped_attr_basename_trie)},\n',
            f')',
        ))

    # ..................{ PERMUTERS                          }..................
    def permute(self) -> 'BeartypeNodeScopeBeforelist':
        '''
        Shallow copy of this scope beforelist, typically called to produce a
        mutable copy of this scope beforelist isolated to a new local (e.g.,
        function) scope being visited by the abstract syntax tree (AST)
        transformer tracking problematic third-party imports.

        Specifically, this shallow copy permutes this scope as follows:

        * The :attr:`.schema_attr_basename_trie` instance variable is
          shallowly copied by calling its :meth:`.ChainMap.new_child` method,
          whose :attr:`.ChainMap.maps` list is then prefixed by a new empty
          dictionary enabling callers to track local imports safely.
        * The :attr:`.module_to_type_to_method_decor_names` instance variable
          is shallowly copied in the same manner.

        Returns
        -------
        BeartypeNodeScopeBeforelist
            Shallow copy of this scope beforelist.
        '''
        # print('Permuting scope beforelist...')

        # Imported attribute name trie unique to this new local scope,
        # initialized to either...
        scoped_attr_basename_trie = (
            # If *NO* transitive parent scope of this scope imported a
            # decorator-hostile decorator, this scope is the first scope in this
            # hierarchy of scopes to require an imported attribute name trie. In
            # this case, instantiate this trie as an empty chain map.
            ChainMap()
            if self.scoped_attr_basename_trie is None else
            # Else, some transitive parent scope of this scope already imported
            # a decorator-hostile decorator and thus required this trie. In this
            # case, enable callers to track imports safely across this new scope
            # by shallowly copying this trie unique to this scope.
            self.scoped_attr_basename_trie.new_child()
        )

        # Create and return a shallow copy of this parent beforelist.
        return BeartypeNodeScopeBeforelist(
            scoped_attr_basename_trie=scoped_attr_basename_trie,

            # Share all remaining data structures of this parent beforelist
            # with this child beforelist.
            schema_attr_basename_trie=self.schema_attr_basename_trie,
            schema_package_names=self.schema_package_names,

            # Avoid uselessly recursively re-validating the contents of these
            # data structures for efficiency.
            is_validate=False,
        )

# ....................{ RAISERS                            }....................
def die_unless_decor_hostile_func_trie(
    # Mandatory parameters.
    schema_attr_basename_trie: BeartypeDecorPlaceTrie,

    # Optional parameters.
    exception_cls: TypeException = BeartypeConfParamException,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed data structure is a valid **decorator
    function beforelist** (i.e., recursive tree structure whose nodes are the
    unqualified basenames of decorator-hostile decorator functions and the
    third-party (sub)packages and (sub)modules transitively defining those
    functions).

    Parameters
    ----------
    schema_attr_basename_trie : BeartypeDecorPlaceTrie
        Decorator function beforelist to be validated.
    exception_cls : type[Exception], default: BeartypeConfParamException
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`.BeartypeConfParamException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this data structure is *not* a valid decorator function beforelist.
    '''
    assert isinstance(exception_cls, type), (
        f'{repr(exception_cls)} not exception type.')
    assert isinstance(exception_prefix, str), (
        f'{repr(exception_prefix)} not string.')

    # If this data structure is *not* a frozen dictionary, raise an exception.
    if not isinstance(schema_attr_basename_trie, FrozenDict):
        raise exception_cls(
            f'{exception_prefix}{repr(schema_attr_basename_trie)} not '
            f'frozen dictionary.'
        )
    # Else, this data structure is a frozen dictionary.

    # If this data structure is *not* a valid decorator function beforelist,
    # raise an exception.
    if not is_decor_hostile_func_trie(schema_attr_basename_trie):
        raise exception_cls(
            f'{exception_prefix}{repr(schema_attr_basename_trie)} not '
            f'decorator-hostile decorator name trie (i.e., '
            f'recursive tree structure satisfying the recursive type hint '
            f'BeartypeDecorPlaceTrie = FrozenDict[str, FrozenDict[str, '
            f'"BeartypeDecorPlaceTrie"]]).'
        )
    # Else, this data structure is a valid decorator function beforelist.

# ....................{ TESTERS                            }....................
def is_decor_hostile_func_trie(
    schema_attr_basename_trie: object) -> TypeIs[BeartypeDecorPlaceTrie]:
    '''
    :data:`True` only if the passed data structure is a valid **decorator
    function beforelist** (i.e., recursive tree structure whose nodes are the
    unqualified basenames of decorator-hostile decorator functions and the
    third-party (sub)packages and (sub)modules transitively defining those
    functions).

    Parameters
    ----------
    schema_attr_basename_trie : BeartypeDecorPlaceTrie
        Decorator function beforelist to be inspected.

    Raises
    ------
    bool
        :data:`True` only if this is a valid decorator function beforelist.
    '''

    #FIXME: *INSUFFICIENT.* To quote the "BeartypeDecorPlaceTrie" hint:
    #    "Note that the root trie is guaranteed to map from strings to *only*
    #    nested frozen dictionaries (rather than to both nested frozen
    #    dictionaries and :data:`None`). Consequently, this hint intentionally
    #    differentiates between matching the root and non-root nesting levels of
    #    this trie."
    #
    #Ergo, we need to expand this validation to:
    #* Non-recursively match the outermost layer to *NOT* map to "None".
    #* Recursively match all non-inner layers to map to
    #  "BeartypeDecorPlaceTrieABC" instances.
    #* Recursively match all inner layers to optionally map to "None".

    # Return true only if...
    return (
        # The passed object is a frozen dictionary *AND*...
        isinstance(schema_attr_basename_trie, FrozenDict) and
        # For each key-value pair of this frozen dictionary...
        all(
            (
                # This key is a string *AND*...
                isinstance(module_name, str) and
                # This value is either...
                (
                    # "None", signifying a terminal leaf node.
                    submodule_or_func_name is None or
                    # A recursively nested frozen dictionary satisfying the same
                    # data structure, signifying a non-terminal stem node.
                    is_decor_hostile_func_trie(submodule_or_func_name)
                )
            )
            for module_name, submodule_or_func_name in (
                schema_attr_basename_trie.items())
        )
    )

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
def make_node_scope_beforelist_global() -> BeartypeNodeScopeBeforelist:
    '''
    Beartype **abstract syntax tree (AST) global scope beforelist** (i.e.,
    low-level dataclass aggregating all metadata required to manage the
    beforelist automating decorator positioning for the global scope of the
    module being recursively visited by the current AST transformer).

    Returns
    -------
    BeartypeNodeScopeBeforelist
        Global scope beforelist of the currently visited module.
    '''

    # Create and return this global scope beforelist.
    return BeartypeNodeScopeBeforelist(
        schema_attr_basename_trie=DECOR_HOSTILE_ATTR_NAME_TRIE)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **assignment statement abstract syntax tree (AST) transformer mixin**
(i.e., low-level superclass instrumenting assignment statements relevant to
runtime type-checking in modules hooked by :mod:`beartype.claw` import hooks).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    AnnAssign,
    Assign,
    Attribute,
    Call,
    Name,
    unparse,
)
from beartype._data.claw.dataclawmagic import BEARTYPE_RAISER_FUNC_NAME
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._data.typing.datatyping import NodeVisitResult
from beartype._util.ast.utilastmake import (
    make_node_call_expr,
    make_node_kwarg,
    make_node_name_load,
    make_node_object_attr_load,
    make_node_str,
)
from beartype._util.text.utiltextansi import color_attr_name

# ....................{ SUBCLASSES                         }....................
class BeartypeNodeTransformerAssignMixin(object):
    '''
    Beartype :pep:`526`-compliant **abstract syntax tree (AST) node
    transformer** (i.e., visitor pattern recursively transforming *all*
    :pep:`526`-compliant annotated variable assignments in the AST tree passed
    to the :meth:`visit` method of the
    :class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass
    also subclassing this mixin).
    '''

    # ..................{ VISITORS ~ pep : 526               }..................
    def visit_Assign(self, node: Assign) -> NodeVisitResult:
        '''
        Track the passed **assignment node** (i.e., node signifying the
        assignment of an attribute) if this node signifies an instantiation of a
        third-party type defining one or more decorator-hostile decorators
        previously imported into a scope of the currently visited module.

        Parameters
        ----------
        node : Assign
            Assignment node to be tracked.

        Returns
        -------
        NodeVisitResult
            This assignment node unmodified.
        '''

        # ..................{ PREAMBLE                       }..................
        # Recursively transform *ALL* child nodes of this parent node.
        self.generic_visit(node)  # type: ignore[attr-defined]

        # ..................{ DECORATOR-HOSTILE              }..................
        # Child node encapsulating the right-hand side (RHS) of this assignment
        # statement, localized purely as a negligible optimization.
        node_source = node.value

        # If this attribute is assigned the result of a call...
        if isinstance(node_source, Call):
            # For each child node encapsulating a left-hand side (LHS) of this
            # assignment statement...
            for node_name_assigned in node.targets:
                # Map the hint annotating the target attribute being assigned to
                # if this call instantiates a third-party type transitively
                # defining decorator-hostile decorators *OR* silently reduce to
                # a noop otherwise (i.e., if this call is *NOT* such an
                # instantiation).
                self.map_node_attr_imported_to_assigned(  # type: ignore[attr-defined]
                    node_name_imported=node_source.func,
                    node_name_assigned=node_name_assigned,
                )
        # Else, this attribute is *NOT* assigned the result of a call.

        # ..................{ RETURN                         }..................
        # Return this node unmodified.
        return node

    # ..................{ VISITORS ~ pep : 526               }..................
    def visit_AnnAssign(self, node: AnnAssign) -> NodeVisitResult:
        '''
        Add a new child node to the passed **annotated assignment node** (i.e.,
        node signifying the assignment of an attribute annotated by a
        :pep:`526`-compliant type hint) inserting a subsequent statement
        following that annotated assignment type-checking that attribute against
        that type hint by passing both to our :func:`beartype.door.is_bearable`
        tester.

        This visitor also additionally track this node if this node signifies an
        instantiation of a third-party type defining one or more
        decorator-hostile decorators previously imported into a scope of the
        currently visited module.

        Design
        ------
        Note that the :class:`.AnnAssign` subclass defines these instance
        variables:

        * ``node.annotation``, a child node describing the PEP-compliant type
          hint annotating this assignment, typically an instance of either:

          * :class:`ast.Name`.
          * :class:`ast.Constant`.

          Note that this node is *not* itself a valid PEP-compliant type hint
          and should *not* be treated as such here or elsewhere.
        * ``node.target``, a child node describing the target attribute assigned
          to by this assignment, guaranteed to be an instance of either:

          * :class:`ast.Name`, in which case this is a **simple assignment**
            (i.e., to a local or global variable). This is the common case in
            which the attribute being assigned to is *NOT* embedded in
            parentheses and thus denotes a simple attribute name rather than a
            full-blown Python expression.
          * :class:`ast.Attribute`, in which case this is an **object
            assignment** (i.e., to an instance or class variable of an object).
          * :class:`ast.Subscript`, in which case this assignment is to the item
            subscripted by an index of a container rather than to that container
            itself.

        * ``node.simple``, an integer :superscript:`sigh` that is either:

          * If ``node.target`` is an :class:`ast.Name` node, 1.
          * Else, 0.

        * ``node.value``, an optional child node defined as either:

          * If this attribute is actually assigned to, a node encapsulating
            the new value assigned to this target attribute.
          * Else, :data:`None`.

        Caveats
        -------
        You may now be thinking to yourself as you wear a bear hat while
        rummaging through this filthy code: "What do you mean, 'if this
        attribute is actually assigned to'? Isn't this attribute necessarily
        assigned to? Isn't that what the 'AnnAssign' subclass means? I mean,
        it's right there in the bloody subclass name: 'AnnAssign', right?
        Clearly, *something* is bloody well being assigned to. Right?"
        Wrong. The name of the :class:`.AnnAssign` subclass was poorly chosen.
        That subclass ambiguously encapsulates both:

        * Annotated variable assignments (e.g., ``muh_attr: int = 42``).
        * Annotated variables *without* assignments (e.g., ``muh_attr: int``).

        Parameters
        ----------
        node : AnnAssign
            Annotated assignment node to be transformed.

        Returns
        -------
        NodeVisitResult
            Either:

            * If this annotated assignment node is *not* **simple** (i.e., the
              attribute being assigned to is embedded in parentheses and thus
              denotes a full-blown Python expression rather than a simple
              attribute name), that same parent node unmodified.
            * If this annotated assignment node is *not* **assigned** (i.e., the
              attribute in question is simply annotated with a type hint rather
              than both annotated with a type hint *and* assigned to), that same
              parent node unmodified.
            * Else, a 2-list comprising both that node and a new adjacent
              :class:`Call` node performing this type-check.

        See Also
        --------
        https://github.com/awf/awfutils
            Third-party Python package whose ``@awfutils.typecheck`` decorator
            implements statement-level :func:`isinstance`-based type-checking in
            a similar manner, strongly inspiring this implementation. Thanks so
            much to Cambridge researcher @awf (Andrew Fitzgibbon) for the
            phenomenal inspiration!
        '''

        # ..................{ PREAMBLE                       }..................
        # Recursively transform *ALL* child nodes of this parent node.
        self.generic_visit(node)  # type: ignore[attr-defined]

        # ..................{ DECORATOR-HOSTILE              }..................
        # Child node encapsulating the hint annotating the target attribute
        # being assigned to, localized purely as a negligible optimization.
        node_hint = node.annotation

        # Child node encapsulating the target attribute being assigned to,
        # localized purely as a negligible optimization.
        node_target = node.target

        # Map the hint annotating the target attribute being assigned to if
        # this hint is the simple name of an isinstanceable third-party type
        # transitively defining decorator-hostile decorators *OR* silently
        # reduce to a noop otherwise (i.e., if this hint is *NOT* such a name).
        self.map_node_attr_imported_to_assigned(  # type: ignore[attr-defined]
            node_name_imported=node_hint,
            node_name_assigned=node_target,
        )

        # ..................{ NOOP                           }..................
        # If either...
        if (
            # It is *NOT* the case that...
            not (
                # This beartype configuration enables type-checking of PEP
                # 526-compliant annotated variable assignments *AND*...
                self._conf.claw_is_pep526 and  # type: ignore[attr-defined]
                # This statement is an assignment (e.g., "muh_var: int = 2")
                # rather than just an unassigned annotation of an attribute
                # (e.g., "muh_var: int").
                node.value
            # Then either this beartype configuration disables type-checking of
            # PEP 526-compliant annotated variable assignments *OR* this
            # statement is just an unassigned annotation of an attribute *OR*...
            ) or
            # This assignment node has one or more parent nodes previously
            # visited by this node transformer *AND* the immediate parent node
            # of this assignment node is a class node, then this assignment node
            # encapsulates a PEP 681-compliant annotated field declaration
            # rather than an PEP 526-compliant annotated variable assignment. In
            # this case, the visit_ClassDef() method defined above has already
            # explicitly decorated the class declaring this annotated field by
            # the @beartype decorator, which then implicitly decorates both this
            # and all other fields of that class by that decorator. For safety
            # and efficiency, avoid needlessly re-decorating this field by the
            # same decorator by simply preserving and returning this node as is.
            #
            # Note, however, that this is *NOT* simply an efficiency concern.
            # This is a significant semantic concern. While a subset of PEP
            # 681-compliant annotated field declarations *ARE* amenable to
            # type-checking by our die_if_unbearable() raiser, still others are
            # absolutely *NOT* amenable to such type-checking. Indeed, in both
            # the average and the worst case, PEP 681-compliant annotated field
            # declarations both supersede and violate PEP 484-compliant typing
            # semantics. Since PEP 681 assumes supremacy over PEP 484 here,
            # @beartype has little to say and much to ignore: e.g.,
            #     from dataclasses import dataclass, field
            #
            #     @dataclass
            #     class MuhDataclass(object):
            #         # This annotated field declaration is safely
            #         # type-checkable by die_if_unbearable(), clearly.
            #         muh_safe_field: int = 0xBABECAFE
            #
            #         # This annotated field declaration is *NOT* safely
            #         # type-checkable by die_if_unbearable(). Clearly, a
            #         # dataclass "field" instance is *NOT* a valid integer and
            #         # thus violates the type hint annotating this field. Since
            #         # PEP 681 standardizes declarations like this as
            #         # semantically valid, @beartype has *NO* alternative but
            #         # to quietly turn a blind eye to what otherwise might be
            #         # considered a type violation.
            #         muh_unsafe_field: int = field(default=0xCAFEBABE)
            self._scopes.is_scope_class  # type: ignore[attr-defined]
        ):
            # Then simply preserve and return this node as is.
            return node
        # Else:
        # * This beartype configuration enables type-checking of PEP
        #   526-compliant annotated variable assignments.
        # * This assignment is simple and assigning to an attribute name.

        # ..................{ LOCALS                         }..................
        # Human-readable label prefixing the exception message raised by our
        # die_if_unbearable() type-checker called below when the value assigned
        # to this variable violates the type hint annotating this variable. For
        # efficiency, we precompute this label at import hook time.
        exception_prefix: str = None  # type: ignore[assignment]

        # Unqualified basename *OR* partially-qualified name of this variable,
        # relative to the current lexical scope.
        var_scoped_name: str = None  # type: ignore[assignment]

        # Child node passing the value newly assigned to this variable by this
        # assignment as the first parameter to die_if_unbearable().
        node_func_arg_pith: AST = None  # type: ignore[assignment]

        # ..................{ PITH                           }..................
        # If this target variable is a simple local or global variable...
        if isinstance(node_target, Name):
            # Unqualified basename of this variable in this lexical scope.
            var_scoped_name = node_target.id

            # Child node accessing this local or global variable.
            #
            # Note that this call effectively shallowly copies this existing
            # "Name" node with context "NODE_CONTEXT_STORE" into a new "Name
            # node with context "NODE_CONTEXT_LOAD". Sadly, the "ast" module
            # provides *NO* means of shallowly copying nodes subject to a
            # trivial modification like this. Ergo, we have *NO* recourse but to
            # do so manually.
            node_func_arg_pith = make_node_name_load(
                name=var_scoped_name, node_sibling=node)
        # Else, this target variable is *NOT* a simple local or global variable.
        #
        # If this target variable is an instance or class variable...
        elif isinstance(node_target, Attribute):
            #FIXME: Insufficient. Attributes can contain arbitrary nested child
            #nodes, including other attributes and/or names. Thankfully, the
            #only reason to even bother attempting to do this is to rigorously
            #sanitize line and column numbers -- which doesn't appear to be
            #particularly necessary or even desirable for dynamically generated
            #code. For now, we simply shallowly reuse the existing "value" node.
            # # Child node referencing the object containing this instance or
            # # class variable (e.g., the "self" in "self.attr: str = 'Attr!'").
            # node_func_arg_pith_obj = Name(
            #     node_target.value.id, ctx=NODE_CONTEXT_LOAD)
            # copy_node_metadata(node_src=node, node_trg=node_func_arg_pith_obj)

            # Child node referencing this instance or class variable.
            #
            # Note that this call effectively shallowly copies this existing
            # "Attribute" node with context "NODE_CONTEXT_STORE" into a new
            # "Attribute" node with context "NODE_CONTEXT_LOAD". Sadly, the
            # "ast" module provides *NO* means of shallowly copying nodes
            # subject to a trivial modification like this. Ergo, we have *NO*
            # recourse but to do so manually.
            node_func_arg_pith = make_node_object_attr_load(
                node_obj=node_target.value,
                attr_name=node_target.attr,
                node_sibling=node,
            )

            # Partially-qualified name of this variable in this lexical scope,
            # defined by "unparsing" this child node. The standard ast.unparse()
            # function "unparses" (i.e., obtains the machine-readable
            # representations of) arbitrary nodes.
            #
            # Note that the parent object of this attribute is described by the
            # external node "node_target.value", encapsulating an arbitrarily
            # complex Python expression. "Unparsing" this expression manually is
            # *ABSOLUTELY* infeasible.
            var_scoped_name = f'{unparse(node_target.value)}.{node_target.attr}'

        #FIXME: Actually, it *WOULD* be useful to type-check "ast.Subscripted"
        #nodes. There's no particular reason *NOT* to, after all. More
        #type-checking can only be a good thing.

        # Else, this target variable is *NOT* an instance or class variable. In
        # this case, this target variable is currently unsupported by this node
        # transformer for automated type-checking. Simply preserve and return
        # this node as is.
        #
        # Examples include:
        # * "ast.Subscripted", in which case this target variable is an item of
        #   a container. It is unclear whether PEP 526 even supports annotated
        #   variable assignments of container items *OR* whether any @beartype
        #   users even annotate variable assignments of container items. Ergo,
        #   this node transformer currently ignores this odd edge case.
        else:
            return node

        # ..................{ CONF                           }..................
        # List of all nodes encapsulating keyword arguments passed to
        # die_if_unbearable(), defaulting to the empty list and thus *NO* such
        # keyword arguments.
        node_func_kwargs = []

        # If the current beartype configuration is *NOT* the default beartype
        # configuration, this configuration is a user-defined beartype
        # configuration which *MUST* be passed as well. In this case...
        if self._conf != BEARTYPE_CONF_DEFAULT:  # type: ignore[attr-defined]
            # Child node encapsulating the passing of this configuration as the
            # "conf" keyword argument to die_if_unbearable().
            node_func_kwarg_conf = self._make_node_keyword_conf(  # type: ignore[attr-defined]
                node_sibling=node)

            # Append this node to the list of all keyword arguments passed to
            # die_if_unbearable().
            node_func_kwargs.append(node_func_kwarg_conf)
        # Else, this configuration is simply the default beartype configuration.
        # In this case, avoid passing that configuration to the beartype
        # decorator for both efficiency and simplicity.

        # ..................{ CONF ~ exception : prefix      }..................
        # If the lexical scope of this parent node is module scope, this node
        # encapsulates a global variable assignment. In this case...
        if self._scopes.is_scope_module:  # type: ignore[attr-defined]
            # Fully-qualified name of this global variable.
            var_name = f'{self._module_name}.{var_scoped_name}'  # type: ignore[attr-defined]

            # Human-readable label prefixing this exception message.
            exception_prefix = f'Global variable "{color_attr_name(var_name)}" '
        # Else, the lexical scope of this parent node is *NOT* module scope.
        # However, by above, this scope is also *NOT* class scope. By
        # elimination, this scope *MUST* thus be a callable scope. In this
        # case...
        else:
            # Fully-qualified name of the callable defining this local variable.
            callable_name = f'{self._scopes[-1].name}()'  # type: ignore[attr-defined]

            # Human-readable label prefixing this exception message.
            exception_prefix = (
                f'Callable {color_attr_name(callable_name)} '
                f'local variable "{color_attr_name(var_scoped_name)}" '
            )
        # print(f'PEP 526 exception_prefix: {exception_prefix}')

        # Child node encapsulating this label as a string literal.
        node_exception_prefix = make_node_str(
            text=exception_prefix, node_sibling=node)

        # Child node encapsulating the passing of this exception prefix as the
        # "exception_prefix" keyword argument to die_if_unbearable().
        node_func_kwarg_exception_prefix = make_node_kwarg(
            kwarg_name='exception_prefix',
            kwarg_value=node_exception_prefix,
            node_sibling=node,
        )

        # Append this node to the list of all keyword arguments passed to
        # die_if_unbearable().
        node_func_kwargs.append(node_func_kwarg_exception_prefix)

        # ..................{ RETURN                         }..................
        # Child node type-checking this newly assigned attribute against the
        # type hint annotating this assignment via our die_if_unbearable()
        # type-checker.
        node_func = make_node_call_expr(
            func_name=BEARTYPE_RAISER_FUNC_NAME,
            nodes_args=[
                # Child node passing the value newly assigned to this
                # attribute by this assignment as the first parameter.
                node_func_arg_pith,
                # Child node passing the type hint annotating this assignment as
                # the second parameter.
                node_hint,
            ],
            nodes_kwargs=node_func_kwargs,
            node_sibling=node,
        )

        # Return a list comprising these two adjacent nodes.
        #
        # Note that order is *EXTREMELY* significant. This order ensures that
        # this attribute is type-checked after being assigned to, as expected.
        return [node, node_func]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **sequence singletons** (i.e., lists and tuples commonly required
throughout this codebase, reducing space and time consumption by preallocating
widely used set-centric objects).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Any,
    List,
    Tuple,
)

# ....................{ LISTS                              }....................
# Note that this exact type annotation is required to avoid mypy complaints. :O
LIST_EMPTY: List[Any] = []
'''
**Empty list singleton.**
'''

# ....................{ TUPLES                             }....................
# Note that this exact type annotation is required to avoid mypy complaints. :O
TUPLE_EMPTY: Tuple[Any, ...] = ()
'''
**Empty tuple singleton.**

Yes, we know exactly what you're thinking: "Why would anyone do this, @leycec?
Why not just directly access the empty tuple singleton as ``()``?" Because
Python insanely requires us to do this under Python >= 3.8 to detect empty
tuples:

.. code-block:: bash

   $ python3.7
   >>> () is ()
   True   # <-- yes, this is good

   $ python3.8
   >>> () is ()
   SyntaxWarning: "is" with a literal. Did you mean "=="?  # <-- WUT
   >>> TUPLE_EMPTY = ()
   >>> TUPLE_EMPTY is TUPLE_EMPTY
   True  # <-- *FACEPALM*
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **exception testers** (i.e., low-level callables introspecting
metadata associated with various types of exceptions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_exception_message_str(exception: Exception) -> bool:
    '''
    :data:`True` only if the message encapsulated by the passed exception is a
    simple string (as is typically but *not* necessarily the case).

    Parameters
    ----------
    exception : Exception
        Exception to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this exception's message is a simple string.
    '''
    assert isinstance(exception, Exception), f'{repr(exception)} not exception.'

    # Return true only if...
    return bool(
        # Exception arguments are a tuple (as is typically but not necessarily
        # the case) *AND*...
        isinstance(exception.args, tuple) and
        # This tuple is non-empty (as is typically but not necessarily the
        # case) *AND*...
        exception.args and
        # The first item of this tuple is a string providing this exception's
        # message (as is typically but *NOT* necessarily the case)...
        isinstance(exception.args[0], str)
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype code-based operator data validation unit tests.**

This submodule unit tests the subset of the public API of the
:mod:`beartype.vale` subpackage defined by the private
:mod:`beartype.vale._is._valeisoper` submodule.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ class : isequal           }....................
def test_api_vale_isequal_pass() -> None:
    '''
    Test successful usage of the :mod:`beartype.vale.IsEqual` factory.
    '''

    # Defer test-specific imports.
    from beartype.vale import IsEqual
    from beartype.vale._core._valecore import BeartypeValidator

    # Arbitrary tuple objects with which to subscript the "IsEqual" class,
    # exercising edge cases in the __class_getitem__() dunder method.
    UNBODIED_JOY = ('Like an unbodied joy', 'whose race is just begun.')

    # Arbitrary hashable non-tuple objects with which to subscript the
    # "IsEqual" class, exercising edge cases in the __class_getitem__() dunder
    # method.
    KEEN_ARROWS = frozenset(
        {'Keen as are the arrows', 'Of that silver sphere,'})

    # Arbitrary non-tuple object we don't particularly care about.
    HARDLY_SEE = ['Until we hardly see,', 'we feel that it is there.']

    # Validator produced by subscripting the "IsEqual" class with exactly one
    # non-tuple object.
    IsUnbodiedJoy = IsEqual[UNBODIED_JOY]
    IsKeenArrows = IsEqual[KEEN_ARROWS]

    # Assert these validators satisfy the expected API.
    assert isinstance(IsUnbodiedJoy, BeartypeValidator)
    assert isinstance(IsKeenArrows, BeartypeValidator)

    # Assert these validators produce the same objects when subscripted by the
    # same arguments (and are thus memoized on subscripted arguments).
    assert IsUnbodiedJoy is IsEqual[UNBODIED_JOY]
    assert IsKeenArrows is IsEqual[KEEN_ARROWS]

    # Assert these validators accept the objects subscripting these validators.
    assert IsUnbodiedJoy.is_valid(UNBODIED_JOY) is True
    assert IsKeenArrows.is_valid(KEEN_ARROWS) is True

    # Assert these validators accept objects *NOT* subscripting these
    # validators but equal to the objects subscripting these validators.
    assert IsUnbodiedJoy.is_valid(UNBODIED_JOY[:]) is True
    assert IsKeenArrows.is_valid(KEEN_ARROWS.copy()) is True

    # Assert these validators reject objects unequal to the objects
    # subscripting these validators.
    assert IsUnbodiedJoy.is_valid(HARDLY_SEE) is False
    assert IsKeenArrows.is_valid(HARDLY_SEE) is False

    # Assert these validators have the expected representation.
    assert repr(UNBODIED_JOY) in repr(IsUnbodiedJoy)
    assert repr(KEEN_ARROWS) in repr(IsKeenArrows)

    # Validator synthesized from the above validators with the domain-specific
    # language (DSL) supported by these validators.
    IsUnbodiedJoyOrKeenArrows = IsUnbodiedJoy | IsKeenArrows

    # Assert this validator performs the expected validation.
    assert IsUnbodiedJoyOrKeenArrows.is_valid(UNBODIED_JOY[:]) is True
    assert IsUnbodiedJoyOrKeenArrows.is_valid(KEEN_ARROWS.copy()) is True
    assert IsUnbodiedJoyOrKeenArrows.is_valid(HARDLY_SEE) is False

    # Assert this validator provides the expected representation.
    assert '|' in repr(IsUnbodiedJoyOrKeenArrows)


def test_api_vale_isequal_fail() -> None:
    '''
    Test unsuccessful usage of the :mod:`beartype.vale.IsEqual` factory.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeValeSubscriptionException
    from beartype.vale import IsEqual
    from pytest import raises

    # Assert that subscripting the "IsEqual" class with the empty tuple raises
    # the expected exception.
    with raises(BeartypeValeSubscriptionException):
        IsEqual[()]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Testing-specific **type hint metadata class hierarchy** (i.e., hierarchy of
classes encapsulating sample type hints instantiated by the
:mod:`beartype_test.a00_unit.data.hint` submodules).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.typing import (
    Iterable,
    Optional,
    Tuple,
    Type,
    TypeVar,
)
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from collections.abc import Iterable as IterableABC

# ....................{ CLASSES ~ hint : [un]satisfied     }....................
class HintPithSatisfiedMetadata(object):
    '''
    **Type hint satisfied pith metadata** (i.e., dataclass whose instance
    variables describe an object satisfying a type hint when either passed as a
    parameter *or* returned as a value annotated by that hint).

    Attributes
    ----------
    pith : object
        Arbitrary object *not* satisfying this hint when either passed as a
        parameter *or* returned as a value annotated by this hint.
    is_context_manager : bool
        If this pith is a **context manager** (i.e., object defining both the
        ``__exit__`` and ``__enter__`` dunder methods required to satisfy the
        context manager protocol), this boolean is either:

        * :data:`True` if callers should preserve this context manager as is
          (e.g., by passing this context manager to the decorated callable).
        * :data:`False` if callers should safely open and close this context
          manager to its context *and* replace this context manager with that
          context (e.g., by passing this context to the decorated callable).

        If this pith is *not* a context manager, this boolean is ignored.
        Defaults to :data:`False`.
    is_pith_factory : bool
        :data:`True` only if this pith is actually a **pith factory** (i.e.,
        callable accepting *no* parameters and dynamically creating and
        returning the value to be used as the desired pith, presumably by
        passing this value to the decorated callable). Defaults to
        :data:`False`.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Mandatory parameters.
        pith: object,

        # Optional parameters.
        is_context_manager: bool = False,
        is_pith_factory: bool = False,
    ) -> None:
        assert isinstance(is_context_manager, bool), (
            f'{repr(is_context_manager)} not boolean.')
        assert isinstance(is_pith_factory, bool), (
            f'{repr(is_pith_factory)} not boolean.')

        # Classify all passed parameters.
        self.pith = pith
        self.is_context_manager = is_context_manager
        self.is_pith_factory = is_pith_factory

    # ..................{ STRINGIFIERS                       }..................
    def __repr__(self) -> str:
        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    pith={repr(self.pith)},',
            f'    is_context_manager={repr(self.is_context_manager)},',
            f'    is_pith_factory={repr(self.is_pith_factory)},',
            f')',
        ))


class HintPithUnsatisfiedMetadata(HintPithSatisfiedMetadata):
    '''
    **Type hint unsatisfied pith metadata** (i.e., dataclass whose instance
    variables describe an object *not* satisfying a type hint when either
    passed as a parameter *or* returned as a value annotated by that hint).

    Attributes
    ----------
    exception_str_match_regexes : Iterable[str]
        Iterable of zero or more r''-style uncompiled regular expression
        strings, each matching a substring of the exception message expected to
        be raised by wrapper functions when either passed or returning this
        ``pith``. Defaults to the empty tuple.
    exception_str_not_match_regexes : Iterable[str]
        Iterable of zero or more r''-style uncompiled regular expression
        strings, each *not* matching a substring of the exception message
        expected to be raised by wrapper functions when either passed or
        returning this ``pith``. Defaults to the empty tuple.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,
        *args,

        # Optional parameters.
        exception_str_match_regexes: Iterable[str] = (),
        exception_str_not_match_regexes: Iterable[str] = (),
        **kwargs
    ) -> None:
        assert isinstance(exception_str_match_regexes, IterableABC), (
            f'{repr(exception_str_match_regexes)} not iterable.')
        assert isinstance(exception_str_not_match_regexes, IterableABC), (
            f'{repr(exception_str_not_match_regexes)} not iterable.')
        assert all(
            isinstance(exception_str_match_regex, str)
            for exception_str_match_regex in exception_str_match_regexes
        ), f'{repr(exception_str_match_regexes)} not iterable of regexes.'
        assert all(
            isinstance(exception_str_not_match_regex, str)
            for exception_str_not_match_regex in (
                exception_str_not_match_regexes)
        ), f'{repr(exception_str_not_match_regexes)} not iterable of regexes.'

        # Initialize our superclass with all variadic parameters.
        super().__init__(*args, **kwargs)

        # Classify all remaining passed parameters.
        self.exception_str_not_match_regexes = exception_str_not_match_regexes

        # Classify the tuple of all r''-style uncompiled regular expression
        # strings appended by the tuple of all mandatory such strings.
        self.exception_str_match_regexes = (
            exception_str_match_regexes +
            _EXCEPTION_STR_MATCH_REGEXES_MANDATORY
        )

    # ..................{ STRINGIFIERS                       }..................
    def __repr__(self) -> str:
        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    pith={repr(self.pith)},',
            f'    is_context_manager={repr(self.is_context_manager)},',
            f'    is_pith_factory={repr(self.is_pith_factory)},',
            f'    exception_str_match_regexes={repr(self.exception_str_match_regexes)},',
            f'    exception_str_not_match_regexes={repr(self.exception_str_not_match_regexes)},',
            f')',
        ))

# ....................{ CLASSES ~ hint : superclass        }....................
class HintNonpepMetadata(object):
    '''
    **PEP-noncompliant type hint metadata** (i.e., dataclass whose instance
    variables describe a type hint that is either PEP-noncompliant or *mostly*
    indistinguishable from a PEP-noncompliant type hint with metadata
    applicable to various testing scenarios).

    Examples of PEP-compliant type hints *mostly* indistinguishable from
    PEP-noncompliant type hints include:

    * :func:`typing.NamedTuple`, a high-level factory function deferring to the
      lower-level :func:`collections.namedtuple` factory function creating and
      returning :class:`tuple` instances annotated by PEP-compliant type hints.
    * :func:`typing.TypedDict`, a high-level factory function creating and
      returning :class:`dict` instances annotated by PEP-compliant type hints.

    Attributes
    ----------
    hint : object
        Type hint to be tested.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all settings configuring type-checking for this type hint).
    is_ignorable : bool
        :data:`True` only if this hint is safely ignorable by the
        :func:`beartype.beartype` decorator. Defaults to :data:`False`.
    is_needs_cls_stack : bool
        :data:`True` only if this hint is **type stack-dependent** (i.e., if
        :mod:`beartype` requires the tuple of all classes lexically declaring
        the class variables or methods annotated by this hint to generate code
        type-checking this hint). Defaults to :data:`False`.
    is_supported : bool
        :data:`True` only if this hint is currently supported by
        the :func:`beartype.beartype` decorator. Defaults to :data:`False`.
    piths_meta : Iterable[HintPithSatisfiedMetadata]
        Iterable of zero or more **(un)satisfied metadata objects** (i.e.,
        :class:`HintPithSatisfiedMetadata` and
        :class:`HintPithUnsatisfiedMetadata` instances), each describing an
        arbitrary object either satisfying or violating this hint when passed as
        a parameter *or* returned as a value annotated by this hint. Defaults to
        the empty tuple.
    warning_type : Optional[Type[Warning]]
        Either:

        * If the :func:`beartype.beartype` decorator unconditionally emits a
          non-fatal warning when this type hint annotates *any* parameter or
          return of *any* callable, the type of that warning.
        * Else, :data:`None`.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,
        *,

        # Mandatory keyword-only parameters.
        hint: object,

        # Optional keyword-only parameters.
        conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
        is_ignorable: bool = False,
        is_needs_cls_stack: bool = False,
        is_supported: bool = True,
        piths_meta: Iterable[HintPithSatisfiedMetadata] = (),
        warning_type: Optional[Type[Warning]] = None,
    ) -> None:

        # Validate passed non-variadic parameters.
        assert isinstance(conf, BeartypeConf), (
            f'{repr(conf)} not configuration.')
        assert isinstance(is_ignorable, bool), (
            f'{repr(is_ignorable)} not bool.')
        assert isinstance(is_needs_cls_stack, bool), (
            f'{repr(is_needs_cls_stack)} not bool.')
        assert isinstance(is_supported, bool), (
            f'{repr(is_supported)} not bool.')
        assert isinstance(piths_meta, IterableABC), (
            f'{repr(piths_meta)} not iterable.')
        assert all(
            isinstance(piths_meta, HintPithSatisfiedMetadata)
            for piths_meta in piths_meta
        ), (
            f'{repr(piths_meta)} not iterable of '
            f'"HintPithSatisfiedMetadata" and '
            f'"HintPithUnsatisfiedMetadata" instances.')
        assert isinstance(warning_type, _NoneTypeOrType), (
            f'{repr(warning_type)} neither class nor "None".')

        # Classify all passed parameters.
        self.hint = hint
        self.conf = conf
        self.is_ignorable = is_ignorable
        self.is_needs_cls_stack = is_needs_cls_stack
        self.is_supported = is_supported
        self.piths_meta = piths_meta
        self.warning_type = warning_type

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    hint={repr(self.hint)},',
            f'    conf={repr(self.conf)},',
            f'    is_ignorable={repr(self.is_ignorable)},',
            f'    is_needs_cls_stack={repr(self.is_needs_cls_stack)},',
            f'    is_supported={repr(self.is_supported)},',
            f'    piths_meta={repr(self.piths_meta)},',
            f'    warning_type={repr(self.warning_type)},',
            f')',
        ))

# ....................{ CLASSES ~ hint : subclass          }....................
class HintPepMetadata(HintNonpepMetadata):
    '''
    **PEP-compliant type hint metadata** (i.e., dataclass whose instance
    variables describe a PEP-compliant type hint with metadata applicable to
    various testing scenarios).

    Attributes
    ----------
    pep_sign : HintSign
        **Sign** (i.e., arbitrary object uniquely identifying this
        PEP-compliant type hint) if this hint is uniquely identified by such a
        sign *or* ``None`` otherwise. Examples of PEP-compliant type hints
        *not* uniquely identified by such attributes include those reducing to
        standard builtins on instantiation such as:

        * :class:`typing.NamedTuple` reducing to :class:`tuple`.
        * :class:`typing.TypedDict` reducing to :class:`dict`.
    is_args : bool, optional
        :data:`True` only if this hint is subscripted by one or more
        **arguments** (i.e., PEP-compliant type hints that are *not* type
        variables) and/or **type variables** (i.e., :class:`typing.TypeVar`
        instances). Defaults to :data:`True` only if the machine-readable
        representation of this hint contains one or more "[" delimiters.
    is_pep585_builtin_subbed : bool, optional
        :data:`True` only if this hint is a :pep:`585`-compliant builtin. If
        :data:`True`, then :attr:`is_type_typing` *must* be :data:`False`.
        Defaults to the negation of :attr:`is_pep585_generic` if
        non-:data:`None` *or* :data:`False` otherwise (i.e., if
        :attr:`is_pep585_generic` is :data:`None`).
    is_pep585_generic : bool, optional
        :data:`True` only if this hint is a :pep:`585`-compliant generic. If
        :data:`True`, then :attr:`is_type_typing` *must* be :data:`False`.
        Defaults to :data:`False`.
    is_typeargs : bool, optional
        :data:`True` only if this hint is subscripted by one or more **type
        variables** (i.e., :class:`typing.TypeVar` instances). Defaults to
        ``bool(typeargs_packed)``.
    is_type_typing : bool, optional
        :data:`True` only if this hint's class is defined by the :mod:`typing`
        module. If ``True``, then :attr:`is_pep585_builtin_subbed` and
        :attr:`is_pep585_generic` *must* both be ``False``. Defaults to
        either:

        * If either :attr:`is_pep585_builtin_subbed` *or*
          :attr:`is_pep585_generic` are :data:`True`, :data:`False`.
        * Else, :data:`True`.
    is_typing : bool, optional
        :data:`True` only if this hint itself is defined by the :mod:`typing`
        module. Defaults to :attr:`is_type_typing`.
    isinstanceable_type : Optional[type]
        **Origin type** (i.e., non-:mod:`typing` class such that *all* objects
        satisfying this hint are instances of this class) originating this hint
        if this hint originates from a non-:mod:`typing` class *or* :data:`None`
        otherwise (i.e., if this hint does *not* originate from such a class).
        Defaults to :data:`None`.
    generic_type : Optional[type]
        Subscripted origin type associated with this hint if any *or*
        :data:`None` otherwise (i.e., if this hint is associated with *no* such
        type). Defaults to either:

        * If this hint is subscripted, :attr:`isinstanceable_type`.
        * Else, :data:`None`.
    typehint_cls : Optional[Type[beartype.door.TypeHint]]
        Concrete :class:`beartype.door.TypeHint` subclass responsible for
        handling this hint if any *or* :data:`None` otherwise (e.g., if the
        :mod:`beartype.door` submodule has yet to support this hint).
    typeargs_packed : Tuple[TypeVar, ...], optional
        Tuple of the zero or more :pep:`484`-compliant **type variables** (i.e.,
        :class:`typing.TypeVar` objects) parametrizing this hint. Defaults to
        the empty tuple, implying this hint to be parametrized by *no* type
        variables.

    All remaining keyword arguments are passed as is to the superclass
    :meth:`HintNonpepMetadata.__init__` method.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,
        *,

        # Mandatory keyword-only parameters.
        pep_sign: 'beartype._data.hint.sign.datahintsigncls.HintSign',

        # Optional keyword-only parameters.
        is_args: Optional[bool] = None,
        is_pep585_builtin_subbed: Optional[bool] = None,
        is_pep585_generic: Optional[bool] = None,
        is_typeargs: bool = None,  # type: ignore[assignment]
        is_type_typing: Optional[bool] = None,
        is_typing: Optional[bool] = None,
        isinstanceable_type: Optional[type] = None,
        generic_type: Optional[type] = None,
        typehint_cls: Optional[Type['beartype.door.TypeHint']] = None,
        typeargs_packed: 'TuplePep484612646TypeArgsPacked' = (),
        **kwargs
    ) -> None:

        # Defer test-specific imports.
        from beartype._data.hint.sign.datahintsigncls import HintSign
        from beartype._util.hint.utilhintget import get_hint_repr
        from beartype._util.hint.pep.proposal.pep484612646 import (
            is_hint_pep484612646_typearg_packed)
        from beartype.door import TypeHint

        # Validate passed non-variadic parameters.
        assert isinstance(pep_sign, HintSign), f'{repr(pep_sign)} not sign.'
        assert isinstance(isinstanceable_type, _NoneTypeOrType), (
            f'{repr(isinstanceable_type)} neither class nor "None".')
        assert isinstance(typeargs_packed, tuple), f'{repr(typeargs_packed)} not tuple.'
        assert all(
            is_hint_pep484612646_typearg_packed(typevar)
            for typevar in typeargs_packed
        ), (
            f'{repr(typeargs_packed)} not tuple of type variables.')

        # Initialize our superclass with all remaining variadic parameters.
        super().__init__(**kwargs)

        # Machine-readable representation of this hint.
        hint_repr = get_hint_repr(self.hint)

        # Conditionally default all unpassed parameters.
            # Default this parameter to true only if the machine-readable
            # representation of this hint contains "[": e.g., "List[str]".
        if is_args is None:
            is_args = '[' in hint_repr
        # Default this parameter to true only if...
        if is_pep585_builtin_subbed is None:
            is_pep585_builtin_subbed = (
                # This hint originates from an origin type *AND*...
                isinstanceable_type is not None and
                # The machine-readable representation of this hint is prefixed
                # by the unqualified name of this origin type (e.g., "list[str]
                # " is prefixed by "list"), suggesting this hint to be a PEP
                # 585-compliant builtin.
                hint_repr.startswith(isinstanceable_type.__name__)
            )
            # print(f'is_pep585_builtin_subbed: {is_pep585_builtin_subbed}')
            # print(f'hint_repr: {hint_repr}')
            # print(f'isinstanceable_type.__name__: {isinstanceable_type.__name__}')
        # Default this parameter to false, because we can't think of anything
        # better.
        if is_pep585_generic is None:
            is_pep585_generic = False
        # Default this parameter to the negation of all PEP 585-compliant
        # boolean parameters. By definition, PEP 585-compliant type hints are
        # *NOT* defined by the "typing" module and vice versa.
        if is_type_typing is None:
            is_type_typing = not (
                is_pep585_builtin_subbed or is_pep585_generic)
        # Default this parameter to true only if this hint is expected to be
        # parametrized by one or more type variables.
        if is_typeargs is None:
            is_typeargs = bool(typeargs_packed)
        # Default this parameter to true only if this hint's class is
        # defined by the "typing" module.
        if is_typing is None:
            is_typing = is_type_typing
        # Default this parameter to this hint's type origin only if this
        # hint is subscripted.
        if generic_type is None:
            generic_type = isinstanceable_type if is_args else None

        # Defer validating parameters defaulting to "None" until *AFTER*
        # initializing these parameters above.
        assert isinstance(is_args, bool), (
            f'{repr(is_args)} not bool.')
        assert isinstance(is_pep585_builtin_subbed, bool), (
            f'{repr(is_pep585_builtin_subbed)} not bool.')
        assert isinstance(is_pep585_generic, bool), (
            f'{repr(is_pep585_generic)} not bool.')
        assert isinstance(is_type_typing, bool), (
            f'{repr(is_type_typing)} not bool.')
        assert isinstance(is_typeargs, bool), (
            f'{repr(is_typeargs)} not bool.')
        assert isinstance(is_typing, bool), (
            f'{repr(is_typing)} not bool.')
        assert isinstance(generic_type, _NoneTypeOrType), (
            f'{repr(generic_type)} neither class nor "None".')
        assert isinstance(generic_type, _NoneTypeOrType), (
            f'{repr(generic_type)} neither class nor "None".')
        assert (
            typehint_cls is None or (
                isinstance(typehint_cls, type) and
                issubclass(typehint_cls, TypeHint),
            )
        ), (
            f'{repr(typehint_cls)} neither '
            f'"beartype.door.TypeHint" subclass nor "None".'
        )

        # Validate that the "is_pep585_builtin_subbed" and "is_type_typing"
        # parameters are *NOT* both true. Note, however, that both can be false
        # (e.g., for PEP 484-compliant user-defined generics).
        assert not (
            (is_pep585_builtin_subbed or is_pep585_generic) and is_type_typing), (
            f'Mutually incompatible boolean parameters '
            f'is_type_typing={repr(is_type_typing)} and either '
            f'is_pep585_builtin_subbed={repr(is_pep585_builtin_subbed)} or '
            f'is_pep585_generic={repr(is_pep585_generic)} enabled.'
        )

        # Classify all passed parameters.
        self.generic_type = generic_type
        self.is_args = is_args
        self.is_pep585_builtin_subbed = is_pep585_builtin_subbed
        self.is_pep585_generic = is_pep585_generic
        self.is_typeargs = is_typeargs
        self.is_type_typing = is_type_typing
        self.is_typing = is_typing
        self.isinstanceable_type = isinstanceable_type
        self.pep_sign = pep_sign
        self.typehint_cls = typehint_cls
        self.typeargs_packed = typeargs_packed

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        return '\n'.join((
            f'{self.__class__.__name__}(',
            f'    hint={repr(self.hint)},',
            f'    conf={repr(self.conf)},',
            f'    pep_sign={repr(self.pep_sign)},',
            f'    typehint_cls={repr(self.typehint_cls)},',
            f'    generic_type={repr(self.generic_type)},',
            f'    isinstanceable_type={repr(self.isinstanceable_type)},',
            f'    is_args={repr(self.is_args)},',
            f'    is_ignorable={repr(self.is_ignorable)},',
            f'    is_needs_cls_stack={repr(self.is_needs_cls_stack)},',
            f'    is_pep585_builtin_subbed={repr(self.is_pep585_builtin_subbed)},',
            f'    is_pep585_generic={repr(self.is_pep585_generic)},',
            f'    is_supported={repr(self.is_supported)},',
            f'    is_typeargs={repr(self.is_typeargs)},',
            f'    is_type_typing={repr(self.is_type_typing)},',
            f'    is_typing={repr(self.is_typing)},',
            f'    piths_meta={repr(self.piths_meta)},',
            f'    typeargs_packed={repr(self.typeargs_packed)},',
            f'    warning_type={repr(self.warning_type)},',
            f')',
        ))

# ....................{ PRIVATE ~ constants                }....................
_EXCEPTION_STR_MATCH_REGEXES_MANDATORY = (
    # Ensure *ALL* exception messages contain the substring "type hint".
    # Exception messages *NOT* containing this substring are overly ambiguous
    # and thus effectively erroneous.
    r'\btype hint\b',
)
'''
Tuple of all **mandatory exception matching regexes** (i.e., r''-style
uncompiled regular expression strings, each unconditionally matching a
substring of the exception message expected to be raised by wrapper functions
when either passed or returning *any* possible pith).
'''


_NoneTypeOrType = (type, type(None))
'''
2-tuple matching both classes and the :data:`None` singleton.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype configuration unit tests.**

This submodule unit tests the subset of the public API of the :mod:`beartype`
package defined by the private :mod:`beartype._conf.confmain` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_conf_dataclass() -> None:
    '''
    Test the public :func:`beartype.BeartypeConf` class.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import (
        BeartypeConf,
        BeartypeDecorPlace,
        BeartypeStrategy,
        BeartypeViolationVerbosity,
        FrozenDict,
    )
    from beartype.roar import (
        BeartypeConfParamException,
        BeartypeCallHintParamViolation,
        BeartypeCallHintReturnViolation,
        BeartypeDoorHintViolation,
    )
    from beartype._conf._confoverrides import _hint_overrides_pep484_tower
    from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
    from beartype._util.utilobject import get_object_type_basename
    from pytest import (
        raises,
        warns,
    )

    # ....................{ CLASSES                        }....................
    class FakeBool(object):
        '''
        Fake boolean class, just because. Look. Just accept it.
        '''

        def __bool__(self) -> bool:
            return False

    # ....................{ LOCALS                         }....................
    # Tuple of all substrings to assert as trivially contained within the string
    # returned by the BeartypeConf.__repr__() dunder method, including:
    # * The unqualified basename of this class.
    # * The unqualified basenames of and all public fields of this class.
    BEAR_CONF_REPR_SUBSTRS = (
        'BeartypeConf',
        'claw_decor_place_func',
        'claw_decor_place_type',
        'claw_is_pep526',
        'claw_skip_package_names',
        'hint_overrides',
        'is_color',
        'is_debug',
        'is_pep484_tower',
        'is_pep557_fields',
        'strategy',
        'violation_door_type',
        'violation_param_type',
        'violation_return_type',
        'violation_type',
        'violation_verbosity',
        'warning_cls_on_decorator_exception',
    )

    # Default (i.e., unparametrized) beartype configuration.
    BEAR_CONF_DEFAULT = BeartypeConf()

    # Non-empty hint overrides mapping one or more arbitrary source type hints
    # to corresponding arbitrary target type hints.
    BEAR_HINT_OVERRIDES_NONEMPTY = FrozenDict({bool: FakeBool})

    # All possible keyword arguments initialized to non-default values with
    # which to instantiate a non-default beartype configuration.
    BEAR_CONF_NONDEFAULT_KWARGS = dict(
        claw_decor_place_func=BeartypeDecorPlace.LAST,
        claw_decor_place_type=BeartypeDecorPlace.FIRST,
        claw_is_pep526=False,
        claw_skip_package_names=('Made_contrast_with', 'the_universe',),
        hint_overrides=BEAR_HINT_OVERRIDES_NONEMPTY,
        is_color=True,
        is_debug=True,
        is_pep484_tower=True,
        is_pep557_fields=True,
        strategy=BeartypeStrategy.Ologn,
        violation_door_type=RuntimeError,
        violation_param_type=TypeError,
        violation_return_type=ValueError,
        violation_type=AttributeError,
        violation_verbosity=BeartypeViolationVerbosity.MINIMAL,
        warning_cls_on_decorator_exception=None,
    )

    # Arbitrary parametrized beartype configuration.
    BEAR_CONF_NONDEFAULT = BeartypeConf(**BEAR_CONF_NONDEFAULT_KWARGS)

    # Arbitrary parametrized beartype configurations setting all possible
    # combinations of a default violation type and another violation type.
    BEAR_CONF_NONDEFAULT_VIOLATION_DOOR = BeartypeConf(
        violation_type=AttributeError, violation_door_type=RuntimeError)
    BEAR_CONF_NONDEFAULT_VIOLATION_PARAM = BeartypeConf(
        violation_type=AttributeError, violation_param_type=TypeError)
    BEAR_CONF_NONDEFAULT_VIOLATION_RETURN = BeartypeConf(
        violation_type=AttributeError, violation_return_type=ValueError)

    # # Tuple of these beartype configurations.
    # BEAR_CONFS = (BEAR_CONF_DEFAULT, BEAR_CONF_NONDEFAULT)

    # ....................{ PASS                           }....................
    # Assert beartype configurations to be self-memoizing across both
    # unparametrized and parametrized type instantiations.
    #
    # Note that the latter explicitly validates that this memoization ignores
    # the order in which parameters are passed.
    assert BeartypeConf() is BeartypeConf()
    assert (
        BeartypeConf(
            claw_decor_place_func=BeartypeDecorPlace.LAST,
            claw_decor_place_type=BeartypeDecorPlace.FIRST,
            claw_is_pep526=False,
            claw_skip_package_names=('Made_contrast_with', 'the_universe',),
            hint_overrides=BEAR_HINT_OVERRIDES_NONEMPTY,
            is_debug=True,
            is_color=True,
            is_pep484_tower=True,
            is_pep557_fields=True,
            strategy=BeartypeStrategy.On,
            violation_door_type=RuntimeError,
            violation_param_type=TypeError,
            violation_return_type=ValueError,
            violation_type=AttributeError,
            violation_verbosity=BeartypeViolationVerbosity.MINIMAL,
            warning_cls_on_decorator_exception=UserWarning,
        ) is
        BeartypeConf(
            warning_cls_on_decorator_exception=UserWarning,
            violation_verbosity=BeartypeViolationVerbosity.MINIMAL,
            violation_type=AttributeError,
            violation_return_type=ValueError,
            violation_param_type=TypeError,
            violation_door_type=RuntimeError,
            strategy=BeartypeStrategy.On,
            is_pep557_fields=True,
            is_pep484_tower=True,
            is_color=True,
            is_debug=True,
            hint_overrides=BEAR_HINT_OVERRIDES_NONEMPTY,
            claw_is_pep526=False,
            claw_skip_package_names=('Made_contrast_with', 'the_universe',),
            claw_decor_place_type=BeartypeDecorPlace.FIRST,
            claw_decor_place_func=BeartypeDecorPlace.LAST,
        )
    )

    # ....................{ PASS ~ properties              }....................
    # Assert that the default configuration contains the expected fields.
    assert BEAR_CONF_DEFAULT.claw_decor_place_func is (
        BeartypeDecorPlace.LAST_BEFORE_DECOR_HOSTILE)
    assert BEAR_CONF_DEFAULT.claw_decor_place_type is (
        BeartypeDecorPlace.LAST)
    assert BEAR_CONF_DEFAULT.claw_is_pep526 is True
    assert BEAR_CONF_DEFAULT.claw_skip_package_names == ()
    assert BEAR_CONF_DEFAULT.hint_overrides is FROZENDICT_EMPTY
    assert BEAR_CONF_DEFAULT.is_color is None
    assert BEAR_CONF_DEFAULT.is_debug is False
    assert BEAR_CONF_DEFAULT.is_pep484_tower is False
    assert BEAR_CONF_DEFAULT.is_pep557_fields is False
    assert BEAR_CONF_DEFAULT.strategy is BeartypeStrategy.O1
    assert BEAR_CONF_DEFAULT.violation_door_type is (
        BeartypeDoorHintViolation)
    assert BEAR_CONF_DEFAULT.violation_param_type is (
        BeartypeCallHintParamViolation)
    assert BEAR_CONF_DEFAULT.violation_return_type is (
        BeartypeCallHintReturnViolation)
    assert BEAR_CONF_DEFAULT.violation_type is None
    assert BEAR_CONF_DEFAULT.violation_verbosity is (
        BeartypeViolationVerbosity.DEFAULT)
    assert BEAR_CONF_DEFAULT.warning_cls_on_decorator_exception is None
    assert BEAR_CONF_DEFAULT._is_warning_cls_on_decorator_exception_set is False

    # Assert that the non-default configuration contains the expected fields.
    assert BEAR_CONF_NONDEFAULT.claw_decor_place_func is (
        BeartypeDecorPlace.LAST)
    assert BEAR_CONF_NONDEFAULT.claw_decor_place_type is (
        BeartypeDecorPlace.FIRST)
    assert BEAR_CONF_NONDEFAULT.claw_is_pep526 is False
    assert BEAR_CONF_NONDEFAULT.claw_skip_package_names == (
        'Made_contrast_with', 'the_universe',)
    assert BEAR_CONF_NONDEFAULT.hint_overrides == (
        BEAR_HINT_OVERRIDES_NONEMPTY | _hint_overrides_pep484_tower())
    assert BEAR_CONF_NONDEFAULT.is_color is True
    assert BEAR_CONF_NONDEFAULT.is_debug is True
    assert BEAR_CONF_NONDEFAULT.is_pep484_tower is True
    assert BEAR_CONF_NONDEFAULT.is_pep557_fields is True
    assert BEAR_CONF_NONDEFAULT.strategy is BeartypeStrategy.Ologn
    assert BEAR_CONF_NONDEFAULT.violation_door_type is RuntimeError
    assert BEAR_CONF_NONDEFAULT.violation_param_type is TypeError
    assert BEAR_CONF_NONDEFAULT.violation_return_type is ValueError
    assert BEAR_CONF_NONDEFAULT.violation_type is AttributeError
    assert BEAR_CONF_NONDEFAULT.violation_verbosity is (
        BeartypeViolationVerbosity.MINIMAL)
    assert BEAR_CONF_NONDEFAULT.warning_cls_on_decorator_exception is None
    assert BEAR_CONF_NONDEFAULT._is_warning_cls_on_decorator_exception_set is (
        True)

    # Assert that the non-default violation-specific configurations contain the
    # expected fields.
    assert BEAR_CONF_NONDEFAULT_VIOLATION_DOOR.violation_type is AttributeError
    assert BEAR_CONF_NONDEFAULT_VIOLATION_DOOR.violation_door_type is (
        RuntimeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_DOOR.violation_param_type is (
        AttributeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_DOOR.violation_return_type is (
        AttributeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_PARAM.violation_type is AttributeError
    assert BEAR_CONF_NONDEFAULT_VIOLATION_PARAM.violation_door_type is (
        AttributeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_PARAM.violation_param_type is (
        TypeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_PARAM.violation_return_type is (
        AttributeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_RETURN.violation_type is (
        AttributeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_RETURN.violation_door_type is (
        AttributeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_RETURN.violation_param_type is (
        AttributeError)
    assert BEAR_CONF_NONDEFAULT_VIOLATION_RETURN.violation_return_type is (
        ValueError)

    # ....................{ PASS ~ equals                  }....................
    # Assert that two differing configurations compare unequal.
    assert BEAR_CONF_DEFAULT != BEAR_CONF_NONDEFAULT

    # Assert that two identical configurations compare equal.
    assert BEAR_CONF_DEFAULT == BeartypeConf()
    assert BEAR_CONF_NONDEFAULT == BeartypeConf(**BEAR_CONF_NONDEFAULT_KWARGS)

    # ....................{ PASS ~ hash                    }....................
    # Assert that two identical configurations hash equal.
    assert hash(BEAR_CONF_DEFAULT) == hash(BeartypeConf())
    assert hash(BEAR_CONF_NONDEFAULT) == hash(
        BeartypeConf(**BEAR_CONF_NONDEFAULT_KWARGS))

    # Assert that two differing configurations hash unequal.
    assert hash(BEAR_CONF_DEFAULT) != hash(BEAR_CONF_NONDEFAULT)

    # ....................{ PASS ~ repr                    }....................
    # Unqualified basename of the class of all beartype configurations.
    BEAR_CONF_BASENAME = get_object_type_basename(BEAR_CONF_DEFAULT)

    # Machine-readable representation of the default configuration.
    BEAR_CONF_DEFAULT_REPR = repr(BEAR_CONF_DEFAULT)

    # Assert that this representation is simply the unqualified basename of the
    # class of this configuration followed by empty parens.
    assert BEAR_CONF_DEFAULT_REPR == f'{BEAR_CONF_BASENAME}()'

    # Machine-readable representation of a non-default configuration.
    BEAR_CONF_NONDEFAULT_REPR = repr(BEAR_CONF_NONDEFAULT)

    # Assert that this representation is prefixed by the unqualified basename of
    # the class of this configuration followed by an opening parens.
    assert BEAR_CONF_NONDEFAULT_REPR.startswith(f'{BEAR_CONF_BASENAME}(')

    # Assert that this representation is suffixed by a closing parens.
    assert BEAR_CONF_NONDEFAULT_REPR.endswith(')')

    # Assert that this representation is *NOT* suffixed by a
    # whitespace-delimited comma followed by a closing parens.
    assert not BEAR_CONF_NONDEFAULT_REPR.endswith(', )')

    # Assert that this representation embeds the names and values of all public
    # non-default fields of this configuration.
    for bear_conf_repr_substr in BEAR_CONF_REPR_SUBSTRS:
        assert bear_conf_repr_substr in BEAR_CONF_NONDEFAULT_REPR

    # ....................{ FAIL ~ raise                   }....................
    # Assert that instantiating a configuration with an invalid parameter raises
    # the expected exception.
    with raises(BeartypeConfParamException):
        BeartypeConf(claw_decor_place_func=(
            "High 'mid the shifting domes of sheeted spray"))
    with raises(BeartypeConfParamException):
        BeartypeConf(claw_decor_place_type=(
            "That canopied his path o'er the waste deep;"))
    with raises(BeartypeConfParamException):
        BeartypeConf(claw_is_pep526=(
            'The fountains mingle with the river'))
    with raises(BeartypeConfParamException):
        BeartypeConf(claw_skip_package_names=('A pine,', 'Rock-rooted,'))
    with raises(BeartypeConfParamException):
        BeartypeConf(claw_skip_package_names=(
            'stretched_athwart', b'the vacancy'))
    with raises(BeartypeConfParamException):
        BeartypeConf(claw_skip_package_names=(
            'Yielding_one.only_response.at.each_pause', ''))
    with raises(BeartypeConfParamException):
        BeartypeConf(claw_skip_package_names=(
            package_name
            for package_name in (
                'Its_swinging_boughs', 'to_each_inconstant_blast',)
        ))
    with raises(BeartypeConfParamException):
        BeartypeConf(hint_overrides=(
            'Wildered, and wan, and panting, she returned.'))
    with raises(BeartypeConfParamException):
        BeartypeConf(is_color=(
            'And many sounds, and much of life and death.'))
    with raises(BeartypeConfParamException):
        BeartypeConf(is_debug=(
            'Interpret, or make felt, or deeply feel.'))
    with raises(BeartypeConfParamException):
        BeartypeConf(is_pep484_tower=(
            'In the calm darkness of the moonless nights,'))
    with raises(BeartypeConfParamException):
        BeartypeConf(is_pep557_fields=(
            "Space region'd with life-air; and barren void;"))
    with raises(BeartypeConfParamException):
        BeartypeConf(strategy=(
            'By all, but which the wise, and great, and good'))
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_door_type=(
            'A vision to the sleep of him who spurned'))
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_door_type=bool)
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_param_type=(
            'His strong heart sunk and sickened with excess'))
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_param_type=str)
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_return_type=(
            'Of love. He reared his shuddering limbs and quelled'))
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_return_type=int)
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_type=(
            'Her choicest gifts. He eagerly pursues'))
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_type=complex)
    with raises(BeartypeConfParamException):
        BeartypeConf(violation_verbosity=(
            'His gasping breath, and spread his arms to meet'))
    with raises(BeartypeConfParamException):
        BeartypeConf(warning_cls_on_decorator_exception=(
            'He lived, he died, he sung, in solitude.'))
    with raises(BeartypeConfParamException):
        BeartypeConf(warning_cls_on_decorator_exception=RuntimeError)

    # Assert that instantiating a configuration with conflicting
    # "is_pep484_tower" and "hint_overrides" parameters raises the expected
    # exception.
    with raises(BeartypeConfParamException):
        BeartypeConf(
            is_pep484_tower=True,
            hint_overrides=FrozenDict({float: complex})
        )
    with raises(BeartypeConfParamException):
        BeartypeConf(
            is_pep484_tower=True,
            hint_overrides=FrozenDict({complex: int})
        )

    # Assert that attempting to modify any public read-only property of this
    # dataclass raises the expected exception.
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.claw_decor_place_func = (
            BeartypeDecorPlace.FIRST)
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.claw_decor_place_type = (
            BeartypeDecorPlace.FIRST)
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.claw_is_pep526 = True
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.claw_skip_package_names = ('q','w','e')
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.hint_overrides = {}
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.is_color = True
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.is_debug = True
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.is_pep484_tower = True
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.strategy = BeartypeStrategy.O0
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.violation_door_type = RuntimeError
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.violation_param_type = TypeError
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.violation_return_type = ValueError
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.violation_type = AttributeError
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.violation_verbosity = (
            BeartypeViolationVerbosity.MINIMAL)
    with raises(AttributeError):
        BEAR_CONF_DEFAULT.warning_cls_on_decorator_exception = None

    # ....................{ FAIL ~ warn                    }....................
    # Assert that instantiating a configuration with a deprecated parameter
    # issues the expected warning.
    with warns(DeprecationWarning):
        BeartypeConf(claw_decoration_position_funcs=BeartypeDecorPlace.FIRST)
    with warns(DeprecationWarning):
        BeartypeConf(claw_decoration_position_types=BeartypeDecorPlace.FIRST)
    with warns(DeprecationWarning):
        BeartypeConf(is_check_pep557=True)

# ....................{ TESTS ~ arg                        }....................
def test_conf_is_color(monkeypatch: 'pytest.MonkeyPatch') -> None:
    '''
    Test the ``is_color`` parameter accepted by the
    :class:`beartype.BeartypeConf` class with respect to the external
    ``${BEARTYPE_IS_COLOR}`` shell environment variable also respected by that
    class, which interact in non-trivial ways.

    Parameters
    ----------
    monkeypatch : MonkeyPatch
        :mod:`pytest` fixture allowing various state associated with the active
        Python process to be temporarily changed for the duration of this test.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import BeartypeConf
    from beartype.roar import (
        BeartypeConfShellVarException,
        BeartypeConfShellVarWarning,
    )
    from beartype._data.os.dataosshell import (
        SHELL_VAR_CONF_IS_COLOR_NAME)
    from pytest import (
        raises,
        warns,
    )

    # ....................{ PASS                           }....................
    # Temporarily enable the ${BEARTYPE_IS_COLOR} environment variable.
    monkeypatch.setenv(SHELL_VAR_CONF_IS_COLOR_NAME, 'True')

    # Non-default beartype configuration masquerading as the default
    # beartype configuration.
    bear_conf = BeartypeConf()

    # Assert that this configuration enabled the "is_color" parameter
    # despite the above instantiation *NOT* passing that parameter.
    assert bear_conf.is_color is True

    # Temporarily disable the ${BEARTYPE_IS_COLOR} environment variable.
    monkeypatch.setenv(SHELL_VAR_CONF_IS_COLOR_NAME, 'False')

    # Non-default beartype configuration masquerading as the default
    # beartype configuration.
    bear_conf = BeartypeConf()

    # Assert that this configuration disabled the "is_color" parameter
    # despite the above instantiation *NOT* passing that parameter.
    assert bear_conf.is_color is False

    # Temporarily nullify the ${BEARTYPE_IS_COLOR} environment variable.
    monkeypatch.setenv(SHELL_VAR_CONF_IS_COLOR_NAME, 'None')

    # Non-default beartype configuration masquerading as the default
    # beartype configuration.
    bear_conf = BeartypeConf()

    # Assert that this configuration nullified the "is_color" parameter
    # despite the above instantiation *NOT* passing that parameter.
    assert bear_conf.is_color is None

    # Non-default beartype configuration explicitly passing the same value for
    # the "is_color" parameter as the ${BEARTYPE_IS_COLOR} environment variable
    # is also currently set to.
    bear_conf = BeartypeConf(is_color=None)

    # Assert that this configuration nullified the "is_color" parameter.
    assert bear_conf.is_color is None

    # ....................{ FAIL                           }....................
    # Assert that attempting to instantiate a beartype configuration when the
    # ${BEARTYPE_IS_COLOR} environment variable is set to a valid string value
    # differing from the valid non-string value passed for the "is_color"
    # parameter emits the expected non-fatal warning.
    with warns(BeartypeConfShellVarWarning):
        # Non-default beartype configuration explicitly passing a differing
        # value for the "is_color" parameter as the ${BEARTYPE_IS_COLOR}
        # environment variable is currently set to.
        bear_conf = BeartypeConf(is_color=True)

        # Assert that this configuration override the passed value of the
        # "is_color" parameter in favour of the corresponding value of the
        # ${BEARTYPE_IS_COLOR} environment variable.
        assert bear_conf.is_color is None

    # Assert that attempting to instantiate a beartype configuration when the
    # ${BEARTYPE_IS_COLOR} environment variable is set to invalid string value
    # raises the expected exception.
    with raises(BeartypeConfShellVarException) as exception_info:
        # Temporarily set ${BEARTYPE_IS_COLOR} to an invalid value.
        monkeypatch.setenv(
            SHELL_VAR_CONF_IS_COLOR_NAME,
            'Strangers have wept to hear his passionate notes,',
        )

        # Attempt to instantiate a beartype configuration.
        BeartypeConf()

    # Assert that this exception message contains an expected substring, whose
    # construction is non-trivial and thus liable to improper construction.
    assert '"True", "False", or "None"' in str(exception_info.value)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator parameter kind** unit tests.

This submodule unit tests the :func:`beartype.beartype` decorator with respect
all explicitly supported **parameter kinds** (e.g., keyword-only,
positional-only, variadic positional, variadic keyword).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip

# ....................{ TESTS ~ name                       }....................
def test_decor_arg_name_fail() -> None:
    '''
    Test unsuccessful usage of the :func:`beartype.beartype` decorator for
    callables accepting one or more **decorator-reserved parameters** (i.e.,
    parameters whose names are reserved for internal use by this decorator).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorParamNameException
    from pytest import raises

    # ....................{ RAISES                         }....................
    # Assert that decorating a callable accepting a reserved parameter name
    # raises the expected exception.
    with raises(BeartypeDecorParamNameException):
        @beartype
        def jokaero(weaponsmith: str, __beartype_func: str) -> str:
            return weaponsmith + __beartype_func

# ....................{ TESTS ~ flexible                   }....................
def test_decor_arg_kind_flex() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable passed two
    **flexible parameters** (i.e., non-variadic positional or keyword
    parameters) annotated with PEP-compliant type hints.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from typing import Union

    # ....................{ CALLABLES                      }....................
    @beartype
    def special_plan(
        finally_gone: Union[list, str],
        finally_done: str,
    ) -> Union[bool, int, str]:
        '''
        Decorated callable to be exercised.
        '''

        return ''.join(finally_gone) + finally_done

    # ....................{ PASSES                         }....................
    # Assert this callable returns the expected value.
    assert special_plan(
        ['When everyone ', 'you have ever loved ', 'is finally gone,'],
        finally_done=(
            'When everything you have ever wanted is finally done with,'),
    ) == (
        'When everyone you have ever loved is finally gone,' +
        'When everything you have ever wanted is finally done with,'
    )


@skip('Currently broken due to known issues in decoration-time type-checking.')
def test_decor_arg_kind_flex_optional() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable passed an
    optional flexible parameter whose default value violates the type hint
    annotating that parameter. Since :func:`beartype.beartype` type-checks
    default values at decoration time (rather than call time as is the common
    case for all other parameter types), this extremely common case warrants
    exhaustive testing.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import (
        BeartypeDecorHintParamDefaultForwardRefWarning,
        BeartypeDecorHintParamDefaultViolation,
    )
    from beartype_test._util.pytroar import raises_uncached
    from pytest import warns

    # ....................{ PASS                           }....................
    # Assert that decorating a callable passed an optional flexible parameter
    # annotated by a forward reference that *CANNOT* be resolved at decoration
    # time emits the expected warning (rather than raising an exception).
    #
    # Note that passing nothing to warns() enables undocumented edge-case
    # behaviour in which warns() captures warnings issued by this block
    # *WITHOUT* validating the types of these warnings. See also:
    #     https://docs.pytest.org/en/4.6.x/warnings.html#recwarn
    with warns() as warnings_info:
        @beartype
        def startled_by(
            # Optional parameter whose default value satisfies its type hint.
            his_own_thoughts: str = (
                'There was no fair fiend near him, not a sight'),
            # Optional parameter whose default value violates its type hint,
            # which is a forward reference that *CANNOT* be resolved at
            # decoration time.
            he_looked_around: 'ThereWasNoFairFiendNearHimNotASight' = (
                'Or sound of awe but in his own deep mind.'),
        ) -> str:
            '''
            Decorated callable to be exercised.
            '''

            return his_own_thoughts + str(he_looked_around)

    # Assert that the above decoration emitted exactly one warning.
    assert len(warnings_info) == 1

    # Warning emitted by the above decoration.
    warning_info = warnings_info[0]

    # Assert that this warning is of the expected type.
    assert warning_info.category is (
        BeartypeDecorHintParamDefaultForwardRefWarning)

    # Warning message.
    warning_message = str(warning_info.message)

    # Assert that this message contains the names of this function, the
    # optional parameter whose default value is uncheckable, and the
    # unresolvable forward reference causing this warning.
    assert 'startled_by' in warning_message
    assert 'he_looked_around' in warning_message
    assert 'ThereWasNoFairFiendNearHimNotASight' in warning_message

    # ....................{ FAIL                           }....................
    # Assert that decorating a callable passed an optional flexible parameter
    # whose default value violates the type hint annotating that parameter
    # raises the expected exception.
    with raises_uncached(BeartypeDecorHintParamDefaultViolation) as (
        exception_info):
        @beartype
        def with_doubtful_smile(
            # Optional parameter whose default value satisfies its type hint.
            mocking_its_own: str = (
                'With doubtful smile mocking its own strange charms.'),
            # Optional parameter whose default value violates its type hint.
            strange_charms: bytes = (
                'Startled by his own thoughts he looked around.'),
        ) -> str:
            '''
            Decorated callable to be exercised.
            '''

            return mocking_its_own + str(strange_charms)

    # Exception message raised by the above decoration of this function.
    exception_message = str(exception_info.value)

    # Assert that this message contains the names of both this function *AND*
    # the optional parameter whose default value violates its type hint.
    assert 'with_doubtful_smile' in exception_message
    assert 'strange_charms' in exception_message


def test_decor_arg_kind_flex_varkw() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable passed a
    mandatory flexible parameter, an optional flexible parameter, and a
    variadic keyword parameter, all annotated with PEP-compliant type hints.

    This test exercises a recent failure in our pre-0.10.0 release cycle:
        https://github.com/beartype/beartype/issues/78
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from typing import Union

    # ....................{ CALLABLES                      }....................
    @beartype
    def one_needs_to_have_a_plan(
        someone_said_who_was_turned: Union[bool, str],
        away_into_the_shadows: Union[float, str] = 'and who i had believed',
        **was_sleeping_or_dead,
    ) -> Union[list, str]:
        '''
        Decorated callable to be exercised.
        '''

        return (
            someone_said_who_was_turned + '\n' +
            away_into_the_shadows + '\n' +
            '\n'.join(was_sleeping_or_dead.values())
        )

    # ....................{ PASS                           }....................
    # Assert this callable returns the expected value.
    assert one_needs_to_have_a_plan(
        someone_said_who_was_turned='imagine he said',
        all_the_flesh_that_is_eaten='the teeth tearing into it',
    ) == '\n'.join((
        'imagine he said',
        'and who i had believed',
        'the teeth tearing into it',
    ))

# ....................{ TESTS ~ keyword                    }....................
def test_decor_arg_kind_kw_unknown_fail() -> None:
    '''
    Test unsuccessful usage of the :func:`beartype.beartype` decorator on
    wrapper functions passed unrecognized keyword parameters.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from pytest import raises

    @beartype
    def tau(kroot: str, vespid: str) -> str:
        '''
        Decorated callable to be exercised.
        '''

        return kroot + vespid

    # Assert that calling this callable with an unrecognized keyword parameter
    # raises the expected exception.
    with raises(TypeError) as exception:
        tau(kroot='Greater Good', nicassar='Dhow')

    # Assert that this exception's message is that raised by the Python
    # interpreter on calling the decorated callable rather than that raised by
    # the wrapper function on type-checking that callable. This message is
    # currently stable across Python versions and thus robustly testable.
    assert str(exception.value).endswith(
        "tau() got an unexpected keyword argument 'nicassar'")

# ....................{ TESTS ~ variadic                   }....................
def test_decor_arg_kind_variadic() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable exercising edge
    cases with respect to variadic positional and keyword parameters.

    Specifically, that callable:

    * Explicitly accepts multiple annotated flexible and keyword-only
      parameters.
    * Implicitly accepts an arbitrary number of excess positional parameters via
      an annotated variadic keyword parameter.
    * Implicitly accepts an arbitrary number of excess keyword parameters via an
      annotated variadic keyword parameter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from pytest import raises

    # ....................{ CALLABLES                      }....................
    @beartype
    def images(
        # Annotated flexible parameter.
        all_the_woven_boughs: int,
        # Annotated flexible parameter.
        above: bytes,
        *,
        # Annotated keyword-only parameter.
        and_each_depending_leaf: float,
        # Annotated variadic keyword parameter.
        **and_every_speck: str
    ) -> int:
        '''
        Arbitrary callable decorated by :func:`beartype.beartype`, exercising
        edge cases with respect to variadic positional and keyword parameters.
        '''

        return (
            all_the_woven_boughs +
            len(above) +
            int(and_each_depending_leaf) +
            len(and_every_speck)
        )

    # ....................{ PASS                           }....................
    # Assert that this callable returns the expected integer when passed a
    # joyous medley of both explicitly and implicitly accepted parameters whose
    # values all satisfy their type hints.
    assert images(
        # Pass at least one explicit flexible parameter positionally.
        len('Of azure sky, darting between their chasms;'),
        # Pass at least one explicit flexible parameter by keyword.
        above=b'Images all the woven boughs above,',
        # Pass at least one explicit keyword-only parameter.
        and_each_depending_leaf=4.17,
        # Pass at least two implicit excess keyword parameters.
        of_azure_sky='darting between their chasms;',
        nor_aught_else='in the liquid mirror laves',
    ) == 83

    # ....................{ FAIL                           }....................
    # Assert that this callable raises the expected exception when passed a
    # hateful chorus of both explicitly and implicitly accepted parameters, at
    # least one of whose values violates its type hint.
    with raises(BeartypeCallHintParamViolation):
        images(
            # Pass at least one explicit flexible parameter positionally.
            len('And each depending leaf, and every speck'),
            # Pass at least one explicit flexible parameter by keyword.
            above=b'Its portraiture, but some inconstant star',
            # Pass at least one explicit keyword-only parameter.
            and_each_depending_leaf=92.303,
            # Pass at least two implicit excess keyword parameters, only the
            # last of which violates its type hint.
            between_one_foliaged_lattice='twinkling fair',
            or_painted_bird=b'sleeping beneath the moon',
        )

# ....................{ TESTS ~ keyword-only               }....................
# Keyword-only keywords require PEP 3102 compliance, which has thankfully been
# available since Python >= 3.0.

def test_decor_arg_kind_kwonly_mixed() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable passed one
    optional keyword-only parameter and one mandatory keyword-only parameter
    (in that non-standard and quite counter-intuitive order), each annotated
    with PEP-compliant type hints.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintViolation
    from beartype.typing import Union
    from beartype_test._util.pytroar import raises_uncached

    @beartype
    def my_own_special_plan(
        *,
        i_listened_to_these_words: Union[dict, str] = (
            'and yet I did not wonder'),
        if_this_creature: Union[set, str],
    ) -> Union[tuple, str]:
        '''
        Decorated callable to be exercised.
        '''

        return i_listened_to_these_words + '\n' + if_this_creature

    # Assert this function returns the expected value.
    assert my_own_special_plan(
        if_this_creature='whom I had thought sleeping or') == '\n'.join((
        'and yet I did not wonder', 'whom I had thought sleeping or'))

    # Assert that calling this callable with invalid parameters raises the
    # expected exception.
    with raises_uncached(BeartypeCallHintViolation):
        my_own_special_plan(
            if_this_creature=b'dead would ever approach his vision')


def test_decor_arg_kind_flex_varpos_kwonly() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable passed a
    flexible parameter, a variadic positional parameter, and a keyword-only
    parameter, all annotated with PEP-compliant type hints.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintViolation
    from beartype.typing import Union
    from beartype_test._util.pytroar import raises_uncached

    # Decorated callable to be exercised.
    @beartype
    def tongue_tasting_its_savour(
        turned_away_into_the_shadows: Union[dict, str],
        *all_the_flesh_that_is_eaten: Union[frozenset, str],
        teeth_tearing_into_it: Union[set, str]
    ) -> Union[tuple, str]:
        return (
            turned_away_into_the_shadows + '\n' +
            '\n'.join(all_the_flesh_that_is_eaten) + '\n' +
            teeth_tearing_into_it
        )

    # Assert this callable returns the expected value.
    assert tongue_tasting_its_savour(
        'When all of your nightmares are for a time obscured',
        'As by a shining brainless beacon',
        'Or a blinding eclipse of the many terrible shapes of this world,',
        'When you are calm and joyful',
        'And finally entirely alone,',
        'Then in a great new darkness',
        teeth_tearing_into_it='You will finally execute your special plan',
    ) == '\n'.join((
        'When all of your nightmares are for a time obscured',
        'As by a shining brainless beacon',
        'Or a blinding eclipse of the many terrible shapes of this world,',
        'When you are calm and joyful',
        'And finally entirely alone,',
        'Then in a great new darkness',
        'You will finally execute your special plan',
    ))

    # Assert that calling this callable with invalid parameters raises the
    # expected exception.
    with raises_uncached(BeartypeCallHintViolation):
        tongue_tasting_its_savour(
            'One needs to have a plan, someone said',
            'Who was turned away into the shadows',
            'And who I had believed was sleeping or dead',
            ['Imagine, he said, all the flesh that is eaten',],
            'The teeth tearing into it',
            'The tongue tasting its savour',
            teeth_tearing_into_it='And the hunger for that taste')

# ....................{ TESTS ~ pep 570                    }....................
# Positional-only keywords require PEP 570 compliance and thus Python >= 3.8.

def test_decor_arg_kind_posonly() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable passed two
    positional-only parameters annotated with PEP-compliant type hints.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintViolation
    from beartype_test.a00_unit.data.pep.data_pep570 import pep570_posonly
    from beartype_test._util.pytroar import raises_uncached

    # Wrapper function type-checking this unchecked function.
    the_taste_and_the_hunger = beartype(pep570_posonly)

    # Assert this function returns the expected value.
    assert the_taste_and_the_hunger(
        'Take away everything as it is') == '\n'.join((
        'Take away everything as it is', 'and the tongue'))

    # Assert that calling this callable with invalid parameters raises the
    # expected exception.
    with raises_uncached(BeartypeCallHintViolation):
        the_taste_and_the_hunger(b'That was my plan')


def test_decor_arg_kind_posonly_flex_varpos_kwonly() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on a callable passed a
    mandatory positional-only parameter, optional positional-only parameter,
    flexible parameter, variadic positional parameter, and keyword-only
    parameter, all annotated with PEP-compliant type hints.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from beartype_test.a00_unit.data.pep.data_pep570 import (
        pep570_posonly_flex_varpos_kwonly)

    # Wrapper function type-checking this unchecked function.
    shining_brainless_beacon = beartype(pep570_posonly_flex_varpos_kwonly)

    # Assert this function returns the expected value.
    assert shining_brainless_beacon(
        'When all of your nightmares are for a time obscured',
        'When you are calm and joyful',
        'And finally entirely alone,',
        'Then in a great new darkness',
        your_special_plan='You will finally execute your special plan',
    ) == '\n'.join((
        'When all of your nightmares are for a time obscured',
        'When you are calm and joyful',
        'And finally entirely alone,',
        'Then in a great new darkness',
        'You will finally execute your special plan',
    ))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :pep:`484`-compliant **type variable** (i.e.,
:class:`typing.TypeVar`) unit tests.

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to :pep:`484`-compliant type variables.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ hint : typevar             }....................
def test_decor_pep484_hint_typevar_cacheunworthy() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on distinct callables annotated
    by **ambiguous type variables** (i.e., distinct type variables bound to
    different types that nonetheless ambiguously share the same names and thus
    the same machine-readable representations), validating that
    :func:`beartype.beartype` correctly preserves type variables as is rather
    than attempting to deduplicate type variables.

    See Also
    --------
    https://github.com/beartype/beartype/issues/458
        Issue exercised by this unit test.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype.typing import (
        Tuple,
        TypeVar,
    )
    from pytest import raises

    # ..................{ CLASSES ~ 1                        }..................
    # First arbitrary class hierarchy.
    class TheDimAndHornedMoonHungLow(object):
        '''
        Arbitrary superclass.
        '''

        pass


    class AndPoured(TheDimAndHornedMoonHungLow):
        '''
        Arbitrary subclass.
        '''

        pass

    # ..................{ CLASSES ~ 2                        }..................
    # Second arbitrary class hierarchy.
    class ASeaOfLustre(object):
        '''
        Arbitrary superclass.
        '''

        pass


    class OnTheHorizonsVerge(ASeaOfLustre):
        '''
        Arbitrary subclass.
        '''

        pass

    # ..................{ LOCALS                             }..................
    # Type variables bound to the above superclasses, intentionally given
    # ambiguous names so as to possibly provoke worst-case behaviour from the
    # @beartype decorator.
    T_TheDimAndHornedMoonHungLow = TypeVar('T', bound=TheDimAndHornedMoonHungLow)
    U_TheDimAndHornedMoonHungLow = TypeVar('U', bound=TheDimAndHornedMoonHungLow)
    T_ASeaOfLustre = TypeVar('T', bound=ASeaOfLustre)
    U_ASeaOfLustre = TypeVar('U', bound=ASeaOfLustre)

    # Type hints parametrized by these type variables.
    Tuple_TheDimAndHornedMoonHungLow = Tuple[
        T_TheDimAndHornedMoonHungLow,
        U_TheDimAndHornedMoonHungLow
    ]
    Tuple_ASeaOfLustre = Tuple[
        T_ASeaOfLustre,
        U_ASeaOfLustre
    ]

    # Arbitrary instances of the above subclasses.
    and_poured = AndPoured()
    on_the_horizons_verge = OnTheHorizonsVerge()

    # 2-tuples intentionally referencing these instances twice.
    tuple_and_poured = (and_poured, and_poured)
    tuple_on_the_horizons_verge = (on_the_horizons_verge, on_the_horizons_verge)

    # ..................{ CALLABLES                          }..................
    @beartype
    def that_overflowed_its_mountains(
        yellow_mist: Tuple_TheDimAndHornedMoonHungLow) -> (
        Tuple_TheDimAndHornedMoonHungLow):
        '''
        Arbitrary callable annotated by type hints defined above.
        '''

        return yellow_mist


    @beartype
    def filled_the_unbounded_atmosphere(
        and_drank: Tuple_ASeaOfLustre) -> (
        Tuple_ASeaOfLustre):
        '''
        Arbitrary callable annotated by type hints defined above.
        '''

        return and_drank

    # ..................{ PASS                               }..................
    # Assert these callables succeed when passed objects satisfying the bounds
    # of the type variables annotating these callables.
    assert that_overflowed_its_mountains(tuple_and_poured) is tuple_and_poured
    assert filled_the_unbounded_atmosphere(tuple_on_the_horizons_verge) is (
        tuple_on_the_horizons_verge)

    # ..................{ FAIL                               }..................
    # Assert these callables fail when passed objects violating the bounds
    # of the type variables annotating these callables.
    with raises(BeartypeCallHintParamViolation):
        that_overflowed_its_mountains(tuple_on_the_horizons_verge)
    with raises(BeartypeCallHintParamViolation):
        filled_the_unbounded_atmosphere(tuple_and_poured)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`646`-compliant **type hint test data.**
'''

# ....................{ FIXTURES                           }....................
def hints_pep646_meta() -> 'List[HintPepMetadata]':
    '''
    List of :pep:`646`-compliant **type hint metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata`
    instances describing test-specific :pep:`646`-compliant sample type hints
    with metadata generically leveraged by various PEP-agnostic unit tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11

    # List of all PEP-specific type hint metadata to be returned.
    hints_pep_meta = []

    # If the active Python interpreter targets Python < 3.11, this interpreter
    # fails to support PEP 646. In this case, return the empty list.
    if not IS_PYTHON_AT_LEAST_3_11:
        return hints_pep_meta
    # Else, the active Python interpreter targets Python >= 3.11 and thus
    # supports PEP 646.

    # ..................{ IMPORTS ~ version                  }..................
    # Defer version-specific imports.
    from beartype.typing import NewType
    from beartype._data.hint.sign.datahintsigns import (
        HintSignPep484585GenericSubbed,
    )
    from beartype_test.a00_unit.data.pep.generic.data_pep646generic import (
        Array,
        DType,
        Shape,
    )
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    # ..................{ LOCALS                             }..................
    # PEP 484-compliant aliases defined by PEP 646's "Summary Examples" section.
    Height = NewType('Height', int)
    Width = NewType('Width', int)

    # ..................{ LISTS                              }..................
    # Add PEP 646-specific test type hints to this list.
    hints_pep_meta.extend((
        # ................{ GENERICS ~ multiple                }................
        # PEP 646-compliant generic subclassing the standard PEP 484-compliant
        # "typing.Generic" superclass parametrized by a PEP 484-compliant type
        # variable followed by a PEP 646-compliant unpacked type variable tuple
        # subscripted by one builtin type and two PEP 484-compliant aliases.
        #
        # This example derives from PEP 646's "Summary Examples" section.
        HintPepMetadata(
            hint=Array[float, Height, Width],
            pep_sign=HintSignPep484585GenericSubbed,
            generic_type=Array,
            is_type_typing=True,
            is_typing=False,
            piths_meta=(
                # Generic instance.
                HintPithSatisfiedMetadata(Array()),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Or the familiar visiting of one'),
            ),
        ),
    ))

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-specific type hint metadata.
    return hints_pep_meta



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Decidedly Object-Oriented Runtime-checking (DOOR) multiprocessing
data** submodule.

This submodule forks Python subprocesses with the standard
:mod:`multiprocessing` package, each of which calls public statement-level
runtime type-checkers defined by the :mod:`beartype.door` subpackage. This logic
leverages the `__main__` dunder global accessible *only* from Python scripts and
*must* thus be isolated to this distinct submodule.
'''

# ....................{ IMPORTS                            }....................
from beartype.door import die_if_unbearable
from beartype.roar import BeartypeDoorHintViolation
from multiprocessing import Pool
from pytest import raises

# ....................{ CALLABLES                          }....................
def and_faster_still(beyond_all_human_speed: str) -> None:
    '''
    Arbitrary function calling the :func:`.die_if_unbearable` raiser.

    To enable the parent Python process to validate that this callable actually
    *was* successfully called, this callable prints the passed parameter as is
    to standard output.

    Note that the :meth:`.Pool.imap` method requires that the first parameter
    passed to that method be a callback. This is that callback.

    Parameters
    ----------
    beyond_all_human_speed : str
        Arbitrary string previously passed as the second parameter to the
        :meth:`.Pool.imap` method called below.
    '''

    # Type-check this string with this raiser.
    die_if_unbearable(beyond_all_human_speed, str)

    # Print this string.
    print(beyond_all_human_speed)

# ....................{ MAIN                               }....................
## If the unqualified basename of this submodule is that of the pseudo-module
# "__main__", then this submodule was actually run as a script via Python's
# standard command-line option "-m". In this case, this is the parent Python
# process forking child Python subprocess (rather than one of those child Python
# subprocesses).
if __name__ == '__main__':
    # Call the above function with a valid parameter, which then prints that
    # parameter to standard output.
    and_faster_still('And faster still, beyond all human speed,')

    # Assert that calling the above function with an invalid parameter from
    # within a forked Python subprocess raises the expected violation.
    with raises(BeartypeDoorHintViolation):
        # With a single-process multiprocessing pool...
        with Pool(processes=1) as pool:
            # Fork a single Python process calling the function defined above.
            tuple(pool.imap(and_faster_still, [42]))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`570` **data submodule.**

This submodule exercises :pep:`570` support implemented in the
:func:`beartype.beartype` decorator by declaring callables accepting one or
more **positional-only parameters** (i.e., parameters that *must* be passed
positionally, syntactically followed in the signatures of their callables by
the :pep:`570`-compliant ``/,`` pseudo-parameter).
'''

# ....................{ IMPORTS                            }....................
from typing import Union

# ....................{ CALLABLES                          }....................
def func_args_2_posonly_mixed(
    before_spreading_his_black_wings: Union[bytearray, str],
    reaching_for_the_skies: Union[bool, str] = 'in this forest',
    /,
) -> Union[list, str]:
    '''
    Arbitrary :pep:`570`-compliant callable passed a mandatory and optional
    positional-only parameter, all annotated with PEP-compliant type hints.
    '''

    return (
        before_spreading_his_black_wings + '\n' + reaching_for_the_skies)


def func_args_10_all_except_flex_mandatory(
    in_solitude_i_wander,
    through_the_vast_enchanted_forest,
    the_surrounding_skies='are one',
    /,
    torn_apart_by='the phenomenon of lightning',
    rain_is_pouring_down='my now shivering shoulders',
    *in_the_rain_my_tears_are_forever_lost,
    the_darkened_oaks_are_my_only_shelter,
    red_leaves_are_blown_by='the wind',
    an_ebony_raven_now_catches='my eye.',
    **sitting_in_calmness,
) -> str:
    '''
    Arbitrary :pep:`570`-compliant callable accepting all possible kinds of
    parameters, including both mandatory and optional variants of these kinds
    except mandatory flexible parameters.

    Since callables cannot by definition accept both optional positional-only
    parameters *and* mandatory flexible parameters, this callable necessarily
    omits the latter in favour of the former.
    '''

    # Arbitrary local variable declared in the body of this callable.
    before_spreading_his_black_wings = 'Reaching for the skies.'
    return before_spreading_his_black_wings



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) type variable
classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for :pep:`484`-compliant :attr:`typing.TypeVar` type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.door._cls.doorhint import TupleTypeHints
from beartype.door._cls.pep.doorpep484604 import UnionTypeHint
# from beartype.roar import BeartypeDoorPepUnsupportedException
from beartype.typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)
from beartype._util.cache.utilcachecall import property_cached

# ....................{ SUBCLASSES                         }....................
class TypeVarTypeHint(UnionTypeHint):
    '''
    **Type variable wrapper** (i.e., high-level object encapsulating a low-level
    :pep:`484`-compliant :attr:`typing.TypeVar` type hint).
    '''

    # ..................{ STATIC                             }..................
    # Squelch false negatives from static type checkers.
    if TYPE_CHECKING:
        _hint: TypeVar

    # ..................{ PROPERTIES                         }..................
    @property
    def is_ignorable(self) -> bool:

        # This type variable is ignorable only if this type variable is either:
        # * Unconstrained by any bounds or constraints (and thus effectively
        #   bound only by the "typing.Any" catch-all).
        # * Bound by an ignorable bound: e.g.,
        #       TypeVar('T', bound=object)
        # * Constrained by one or more ignorable constraints. Since constraints
        #   effectively build a union over those constraints, even a single
        #   ignorable constraint suffices to render the entire type variable
        #   ignorable: e.g.,
        #       TypeVar('T', object)
        return self._is_args_ignorable

    # ..................{ PRIVATE ~ properties               }..................
    #FIXME: *HMM.* We should arguably just define the _make_args() factory
    #method instead. That implementation would become quite a bit simpler as
    #well as generalize to cover more use cases. By defining this method,
    #"self._args" and "self._args_wrapped_tuple" are now desynchronized. *sigh*

    @property  # type: ignore[misc]
    @property_cached
    def _args_wrapped_tuple(self) -> TupleTypeHints:

        #FIXME: Support covariance and contravariance, please. We don't
        #particularly care about either at the moment. Moreover, runtime type
        #checkers were *NEVER* expected to support either -- although we
        #eventually intend to do so. For now, raising a fatal exception here
        #would seem to be extreme overkill. Doing nothing is (probably) better
        #than doing something reckless and wild.
        # # Human-readable string describing the variance of this type variable if
        # # any *OR* "None" otherwise (i.e., if this type variable is invariant).
        # variance_str = None
        # if self._hint.__covariant__:
        #     variance_str = 'covariant'
        # elif self._hint.__contravariant__:
        #     variance_str = 'contravariant'
        #
        # # If this type variable is variant, raise an exception.
        # if variance_str:
        #     raise BeartypeDoorPepUnsupportedException(
        #         f'Type hint {repr(self._hint)} '
        #         f'variance "{variance_str}" currently unsupported.'
        #     )
        # # Else, this type variable is invariant.

        # TypeVars may only be bound or constrained, but not both. The
        # difference between the two has semantic meaning for static type
        # checkers but relatively little meaning for us. Ultimately, we're only
        # concerned with the set of compatible types present in either the bound
        # or the constraints. So, we treat a type variable as a union of its
        # constraints or bound. See also:
        #     https://docs.python.org/3/library/typing.html#typing.TypeVar

        # If this type variable is bounded, return the 1-tuple containing only
        # this wrapped bound.
        if self._hint.__bound__ is not None:
            return (TypeHint(self._hint.__bound__),)
        # Else, this type variable is unbounded.
        #
        # If this type variable is constrained, return the n-tuple containing
        # each of these wrapped constraints.
        elif self._hint.__constraints__:
            return tuple(TypeHint(t) for t in self._hint.__constraints__)
        # Else, this type variable is unconstrained.

        #FIXME: Consider globalizing this as a private constant for efficiency.
        # Return the 1-tuple containing only the "typing.Any" catch-all. Why?
        # Because an unconstrained and unbounded type variable is semantically
        # equivalent to a type variable bounded by "typing.Any".
        return (TypeHint(Any),)  # pyright: ignore



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **sample callable** submodule.

This submodule predefines sample pure-Python callables exercising known edge
cases on behalf of higher-level unit test submodules.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Callable
from functools import wraps

# ....................{ DECORATORS                         }....................
def decorator(func: Callable) -> Callable:
    '''
    **Identity decorator** (i.e., decorator returning the passed callable
    unmodified).

    This decorator enables logic elsewhere to exercise the
    :mod:`beartype.beartype` decorator with respect to nested callables
    decorated by both the :mod:`beartype.beartype` decorator and one or more
    decorators *not* the :mod:`beartype.beartype` decorator.
    '''

    return func

# ....................{ DECORATORS ~ hostile               }....................
class DecoratorHostileWrapper(object):
    '''
    **Decorator-hostile wrapper** (i.e., arbitrary uncallable type defining a
    non-standard :meth:`.invoke` method calling the decorated callable or type
    against which this wrapper was instantiated by the parent
    :func:`.decorator_hostile` decorator).

    Attributes
    ----------
    _wrappee : Callable
        **Wrappee** (i.e., callable or type originally decorated by the parent
        :func:`.decorator_hostile` decorator encapsulating that callable or type
        with this wrapper instance).
    '''

    def __init__(self, wrappee: Callable) -> None:
        '''
        Initialize this decorator-hostile wrapper with the passed decorated
        callable or type.
        '''

        # Classify the passed wrappee intelligently, such that...
        self._wrappee = (
            # If this wrappee is actually yet another instance of this same
            # decorator-hostile wrapper class, unwrap this wrapper to the
            # wrappee previously encapsulated by this wrapper.
            wrappee._wrappee
            if isinstance(wrappee, DecoratorHostileWrapper) else
            # Else, this wrappee is presumably a callable or type. In this case,
            # classify this wrappee as is.
            wrappee
        )


    def invoke(self, *args, **kwargs) -> object:
        '''
        Call the decorated callable or type wrapped by this decorator-hostile
        wrapper with all passed positional and keyword parameters.
        '''

        # Defer to the decorated callable or type.
        return self._wrappee(*args, **kwargs)


def decorator_hostile(func: Callable) -> DecoratorHostileWrapper:
    '''
    **Decorator-hostile decorator** (i.e., decorator that prevents other
    decorators from being applied after this decorator is applied to the
    decorated callable or type).

    This decorator returns an arbitrary uncallable object of an arbitrary type
    defining an ``invoke()`` method calling the decorated callable or type.
    This decorator thus "wraps" the decorated callable or type with a
    non-standard calling convention rather than the standard ``__wrapped__``
    dunder attribute defined by the standard :func:`functools.wraps` wrapper
    employed by standard decorators. Doing so prevents other decorators
    (including :func:`beartype.beartype` itself) from unwrapping the decorated
    callable or type from the uncallable object returned by this decorator,
    which then prevents those other decorators from introspecting the decorated
    signature. In short, this decorator is trivially decorator-hostile.

    This decorator intentionally mimics the action of real-world
    decorator-hostile decorators, including:

    * The third-party :func:`langchain_core.runnables.core` decorator function.
    * The third-party :func:`fastmcp.FastMCP.tool` decorator method.
    '''

    # Create and return a decorator-hostile wrapper wrapping this object.
    return DecoratorHostileWrapper(func)


class DecoratorHostileType(object):
    '''
    **Decorator-hostile type** (i.e., arbitrary class defining one or more
    decorator-hostile methods).
    '''

    def __init__(self) -> None:
        '''
        Initialize this decorator-hostile instance.
        '''

        pass


    def decorator_hostile_method(
        self, func: Callable) -> DecoratorHostileWrapper:
        '''
        **Decorator-hostile decorator method** (i.e., decorator implemented as a
        bound instance method preventing other decorators from being applied
        after this decorator is applied to the decorated callable or type).

        See Also
        --------
        :func:`.decorator_hostile`
            Further details.
        '''

        # Create and return a decorator-hostile wrapper wrapping this object.
        return DecoratorHostileWrapper(func)

# ....................{ DECORATORS ~ (non)isomorphic       }....................
def decorator_isomorphic(func: Callable) -> Callable:
    '''
    **Isomorphic decorator** (i.e., function returning an isomorphic decorator
    closure transparently preserving both the positions and types of all
    parameters passed to the passed callable).
    '''

    @wraps(func)
    def _closure_isomorphic(*args, **kwargs):
        '''
        **Isomorphic decorator closure** (i.e., closure transparently
        preserving both the positions and types of all parameters and returns
        passed to the decorated callable).
        '''

        return func(*args, **kwargs)

    # Return this closure.
    return _closure_isomorphic


def decorator_nonisomorphic(func):
    '''
    **Non-isomorphic decorator** (i.e., function returning a non-isomorphic
    decorator closure destroying the positions and/or types of one or more
    parameters passed to the passed callable).
    '''

    @wraps(func)
    def _closure_nonisomorphic():
        '''
        **Non-isomorphic decorator closure** (i.e., closure destroying the
        positions and/or types of one or more parameters passed to the decorated
        callable).
        '''

        return func()

    # Return this closure.
    return _closure_nonisomorphic




#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`557`-compliant **type hint reducers** (i.e., low-level
low-level callables converting higher-level type hints created by subscripting
the :obj:`dataclasses.InitVar` type hint factory to lower-level type hints more
readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatypingport import Hint
from beartype._util.hint.pep.proposal.pep557 import get_hint_pep557_initvar_arg

# ....................{ REDUCERS                           }....................
def reduce_hint_pep557_initvar(hint: Hint, exception_prefix: str) -> Hint:
    '''
    Reduce the passed :pep:`557`-compliant **dataclass initialization-only
    instance variable type hint** (i.e., subscription of the
    :obj:`dataclasses.InitVar` type hint factory) to the child type hint
    subscripting this parent hint -- which is otherwise functionally useless
    from the admittedly narrow perspective of runtime type-checking.

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type variable to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    Hint
        Lower-level type hint currently supported by :mod:`beartype`.
    '''

    # Reduce this "typing.InitVar[{hint}]" type hint to merely "{hint}".
    return get_hint_pep557_initvar_arg(
        hint=hint, exception_prefix=exception_prefix)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype :pep:`544` **type hint utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep544` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep544_protocol() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep544.is_hint_pep544_protocol`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from abc import abstractmethod
    from beartype._util.hint.pep.proposal.pep544 import (
        is_hint_pep544_protocol)
    from beartype_test.a00_unit.data.data_type import (
        TYPES_BUILTIN,
        Class,
    )

    # Intentionally import @beartype-accelerated protocols.
    from beartype.typing import (
        Protocol,
        runtime_checkable,
    )

    # Intentionally import @beartype-unaccelerated protocols.
    from typing import (
        SupportsAbs,
        SupportsBytes,
        SupportsComplex,
        SupportsFloat,
        SupportsInt,
        SupportsRound,
        Union,
    )

    # ....................{ LOCALS                         }....................
    @runtime_checkable
    class ProtocolCustomUntypevared(Protocol):
        '''
        User-defined protocol parametrized by *NO* type variables declaring
        arbitrary concrete and abstract methods.
        '''

        def alpha(self) -> str:
            return 'Of a Spicily sated'

        @abstractmethod
        def omega(self) -> str: pass

    # Set of all PEP 544-compliant "typing" protocols.
    TYPING_PROTOCOLS = {
        SupportsAbs,
        SupportsBytes,
        SupportsComplex,
        SupportsFloat,
        SupportsInt,
        SupportsRound,
    }

    # ....................{ PASS                           }....................
    # Assert this tester rejects builtin types erroneously presenting themselves
    # as PEP 544-compliant protocols under Python >= 3.8.
    assert is_hint_pep544_protocol(int) is False
    assert is_hint_pep544_protocol(str) is False
    assert is_hint_pep544_protocol(type(None)) is False

    # Assert this tester rejects a user-defined type.
    assert is_hint_pep544_protocol(Class) is False
    # class Yam(object): pass
    # from typing import Protocol
    # assert not issubclass(Yam, Protocol)
    # assert is_hint_pep544_protocol(Yam) is False

    # Assert this tester accepts all standard PEP 544-compliant protocols.
    for typing_protocol in TYPING_PROTOCOLS:
        assert is_hint_pep544_protocol(typing_protocol) is True

    # Assert this tester rejects all builtin types. For unknown reasons, some
    # but *NOT* all builtin types (e.g., "int") erroneously present themselves
    # to be PEP 544-compliant protocols. *sigh*
    for class_builtin in TYPES_BUILTIN:
        assert is_hint_pep544_protocol(class_builtin) is False

    # Assert this tester rejects standard type hints in either case.
    assert is_hint_pep544_protocol(Union[int, str]) is False

    # Assert this tester accepts a standardized protocol.
    assert is_hint_pep544_protocol(SupportsInt) is True

    # Assert this tester accepts a user-defined protocol.
    assert is_hint_pep544_protocol(ProtocolCustomUntypevared) is True


def test_is_hint_pep544_io_generic() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep544.is_hint_pep484_generic_io`
    tester.
    '''

    # Defer test-specific imports.
    from beartype.typing import Union
    from beartype._data.cls.datacls import TYPES_PEP484_GENERIC_IO
    from beartype._util.hint.pep.proposal.pep544 import (
        is_hint_pep484_generic_io)

    # Assert this tester accepts all standard PEP 484-compliant IO generics.
    for pep484_generic_io in TYPES_PEP484_GENERIC_IO:
        assert is_hint_pep484_generic_io(pep484_generic_io) is True

    # Assert this tester rejects standard type hints in either case.
    assert is_hint_pep484_generic_io(Union[int, str]) is False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`, :pep:`612`, and :pep:`644` **type parameter utility**
unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484612646` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep484612646_typearg_packed_name() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep484612646.get_hint_pep484612646_typearg_packed_name`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484612646Exception
    from beartype._data.typing.datatyping import T
    from beartype._util.hint.pep.proposal.pep484612646 import (
        get_hint_pep484612646_typearg_packed_name)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
    from beartype_test.a00_unit.data.pep.data_pep612 import P
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this getter passed a PEP 484-compliant type variable returns the
    # name of this type variable.
    assert get_hint_pep484612646_typearg_packed_name(T) == 'T'

    # Assert this getter passed a PEP 612-compliant parameter specification
    # returns the name of this parameter specification.
    assert get_hint_pep484612646_typearg_packed_name(P) == 'P'

    # If the active Python interpreter targets Python >= 3.11 and thus supports
    # PEP 646...
    if IS_PYTHON_AT_LEAST_3_11:
        # Defer version-specific imports.
        from beartype_test.a00_unit.data.pep.data_pep646 import Ts

        # Assert this getter passed a PEP 646-compliant type variable tuple
        # returns the name of this type variable tuple.
        assert get_hint_pep484612646_typearg_packed_name(Ts) == 'Ts'

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an object
    # that is *NOT* a type parameter.
    with raises(BeartypeDecorHintPep484612646Exception):
        get_hint_pep484612646_typearg_packed_name(
            'As with a palsied tongue, and while his beard')

# ....................{ TESTS ~ packer                     }....................
def test_pack_hint_pep484612646_typearg_unpacked() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep484612646.pack_hint_pep484612646_typearg_unpacked`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484612646Exception
    from beartype._data.typing.datatyping import T
    from beartype._util.hint.pep.proposal.pep484612646 import (
        pack_hint_pep484612646_typearg_unpacked)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this function preserves a PEP 484-compliant type variable as is.
    assert pack_hint_pep484612646_typearg_unpacked(T) is T

    #FIXME: Actually test parameter specification packing. We can't be
    #bothered at the moment -- mostly because nobody cares about PEP 612 in
    #the context of runtime type-checking. Also, this is obviously wrong. We're
    #testing unpacking here. So, we'd need to define a new pair of
    #"P_unpacked_prefix" and "P_unpacked_subbed" hints. *sigh*
    # # Defer version-specific imports.
    # from beartype_test.a00_unit.data.pep.data_pep612 import (
    #     P,
    #     P_unpacked_prefix,
    #     P_unpacked_subbed,
    # )
    #
    # # Assert this getter passed a PEP 612-compliant parameter specification
    # # returns the name of this parameter specification.
    # assert pack_hint_pep484612646_typearg_unpacked(P_unpacked_prefix) is P
    # assert pack_hint_pep484612646_typearg_unpacked(P_unpacked_subbed) is P

    # If the active Python interpreter targets Python >= 3.11 and thus
    # supports PEP 646...
    if IS_PYTHON_AT_LEAST_3_11:
        # Defer version-specific imports.
        from beartype_test.a00_unit.data.pep.data_pep646 import (
            Ts,
            Ts_unpacked_prefix,
            Ts_unpacked_subbed,
        )

        # Assert this function packs a PEP 646-compliant unpacked type
        # variable tuple to that type variable tuple regardless of whether
        # that type variable tuple was unpacked in either:
        # * Prefix form (e.g., as "*Ts").
        # * In subscripted form (e.g., as "typing.Unpack[Ts]").
        assert pack_hint_pep484612646_typearg_unpacked(
            Ts_unpacked_prefix) is Ts
        assert pack_hint_pep484612646_typearg_unpacked(
            Ts_unpacked_subbed) is Ts

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an object
    # that is *NOT* a type parameter.
    with raises(BeartypeDecorHintPep484612646Exception):
        pack_hint_pep484612646_typearg_unpacked(
            "Bastion'd with pyramids of glowing gold,")
